/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: argfile.h 2179 2011-06-24 09:03:24Z ahuesgen $
 */

#ifndef ARGFILE_INCLUDED
#define ARGFILE_INCLUDED

#include "../shared/misc_defines.h"
#include "wsfile.h"

#include <QString>
#include <QHash>
#include <QList>

namespace vhist
{

typedef QPair<QString, QString> UserAttribute;

/**
 * A class containing all information stored in a VHIST argfile.
 * The information is stored in a structured way. The content of this
 * class can be serialized to an argfile using toCommandline(). To
 * generate an ArgFile object from a given argfile, use
 * vhist::ArgfileReader.
 *
 * @see vhist::ArgfileReader
 */
class VHIST_EXTERN ArgFile
{
    public:
        /**
         * Specifies the amount of output generated by \a vhistadd
         * when executing this argfile.
         */
        enum Verbosity
        {
            /// \a vhistadd will only generate error messages.
            QUIET,
            /// \a vhistadd will generate a normal amount of output.
            NORMAL,
            /// \a vhistadd will generate a large amount of output.
            VERBOSE
        };
    
    public: // construction, destruction
        ArgFile();
        ArgFile(const QString& outVhistfile, const QString& rootfile = QString());
        ArgFile(const ArgFile& rhs);
        ~ArgFile();
    
    public: // main functionality
        QString toCommandline(bool supportVariables = true) const;
    
    public: // getter and setter
        void setPretend(bool pretend);
        bool pretend() const;
        void setDebugMode(bool debug);
        bool debugMode() const;
        void setVerbosity(Verbosity verbosity);
        Verbosity verbosity() const;
        void setShowHelp(bool showHelp);
        bool showHelp() const;
        void setShowVersion(bool showVersion);
        bool showVersion() const;
        
        void setRootfile(const QString& rootfile);
        QString rootfile() const;
        void setRootIsOptional(bool optional);
        bool rootIsOptional() const;
        
        void setTitle(const QString& title);
        QString title() const;
        void setDescription(const QString& description);
        QString description() const;
        void setComment(const QString& comment);
        QString comment() const;
        void setTool(const QString& tool);
        QString tool() const;
        void setToolpath(const QString& toolpath);
        QString toolpath() const;
        void setHost(const QString& host);
        QString host() const;
        void setUser(const QString& user);
        QString user() const;
        void setCommand(const QString& command);
        QString command() const;
        
        void setPdfTitle(const QString& pdfTitle);
        QString pdfTitle() const;
        void setPdfProducer(const QString& pdfProducer);
        QString pdfProducer() const;
        void setPdfCreator(const QString& pdfCreator);
        QString pdfCreator() const;
        void setPdfAuthor(const QString& pdfAuthor);
        QString pdfAuthor() const;
        void setPdfKeywords(const QString& pdfKeywords);
        QString pdfKeywords() const;
        void setPdfSubject(const QString& pdfSubject);
        QString pdfSubject() const;
        
        void setReadme(const QString& readmeFile);
        QString readme() const;
        void setFirstpage(const QString& firstPageFile);
        QString firstpage() const;

        void clearVariables();
        void clearRootVhistfiles();

	public: // in-/outfile methods
        WSFile* lastFile();
        WSFile* addFile(const WSFile& file);
        int numFiles() const;
        WSFile* getFile(int index);
        WSFile getConstFile(int index) const;
        void clearFilelist();
        void removeFile(int index);
    
    public: // user attribute methods
        void addUserAttribute(const QString& key, const QString& value);
        void clearUserAttributes();
        QStringList userAttributeKeys() const;
        QString userAttribute(const QString& key) const;
    
    public: // output vhistfile methods
        void addOutputVhistfile(const QString& filename);
        void clearOutputVhistfiles();
        int numOutputVhistfiles() const;
        QString outputVhistfile(int index) const;
		QString getUndoWorkflowStep() const;
    
    private: // list/hash options
        QHash<QString, QString> m_userAttributes;
        QStringList m_userAttributesOrder;
        QList<WSFile> m_files;
        QStringList m_outVhistfiles;
    
    private: // misc options
        Verbosity m_verbosity;
        bool m_rootIsOptional;
        bool m_pretend;
        bool m_debugMode;
        bool m_showHelp;
        bool m_showVersion;
        
        QString m_firstpage;
        QString m_readme;
        
        QString m_rootfile;
    
    private: // -s options
        QString m_title;
        QString m_description;
        QString m_comment;
        QString m_tool;
        QString m_toolpath;
        QString m_host;
        QString m_user;
        QString m_command;
    
    private: // -d options
        QString m_pdfTitle;
        QString m_pdfProducer;
        QString m_pdfCreator;
        QString m_pdfAuthor;
        QString m_pdfKeywords;
        QString m_pdfSubject;
};

} // namespace

#endif // ARGFILE_INCLUDED

