/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

$Id: wsfile.h 2179 2011-06-24 09:03:24Z ahuesgen $
*/

#ifndef FILE_INCLUDED
#define FILE_INCLUDED

#include "../shared/misc_defines.h"

#include <QString>
#include <QDir>
#include <QList>
#include <QHash>


namespace vhist
{

/** 
 * This class contains the description of a single
 * input/output file specified in an argfile.
 */
class VHIST_EXTERN WSFile
{
    public:
        /**
         * The purpose of a file inside one workflowstep.
         */
        enum FilePurpose
        {
            /// The file is an input file.
            INPUT_FILE,
            /// The file is an output file.
            OUTPUT_FILE
        };
    
    public: // construction and destruction
        WSFile();
        WSFile(const WSFile& rhs);
    
    public: // comparison
        bool operator==(const WSFile& rhs);
        bool operator!=(const WSFile& rhs);
    
    public: // main functionality
        QString toCommandline(bool supportVariables = true);
    
    public: // getter/setter
        QString filetype() const;
        void setFiletype(const QString& filetype);
        
        QString filename() const;
        void setFilename(const QString& filename);
        
        QString filepath() const;
        void setFilepath(const QString& filepath);
        
        void setFilenameAndPath (const QString& file);
        
        bool optional() const;
        void setOptional(bool optional);
        
        bool compress() const;
        void setCompress(bool compress);
        
        bool inputFile() const;
        void setInputFile(bool isInputFile);
        
        FilePurpose filePurpose() const;
        void setFilePurpose(const FilePurpose& purpose);
        
        QString description() const;
        void setDescription(const QString& description);
        
        QString comment() const;
        void setComment(const QString& comment);
        
        bool embed() const;
        void setEmbed(bool embed);
        
        bool autoMD5() const;
        void setAutoMD5(bool compress);
        
        QString md5File() const;
        void setMD5File(const QString& file);
	
        bool previewWS() const;
        void setPreviewWS(bool previewWS);
	
        bool preview() const;
        void setPreview(bool preview);
	
        bool thumbnail() const;
        void setThumbnail(bool thumbnail);
        
        bool thumbnailOnly() const;
        void setThumbnailOnly(bool thumbnailOnly);
        
        void clearUserAttributes();
        void addUserAttribute(const QString& key, const QString& value);
        QStringList userAttributeKeys() const;
        QString userAttribute(const QString& key) const;
    
    private:
        QString m_filetype;
        QString m_filename;
        QString m_filepath;
        QString m_md5File;
        
        bool m_md5;
        bool m_embed;
        bool m_optional;
        bool m_compress;
        bool m_previewws;
        bool m_preview;
        bool m_thumbnail;
        bool m_thumbnailOnly;
        
        FilePurpose m_filePurpose;
        QString m_description;
        QString m_comment;
        QHash<QString, QString> m_userAttributes;
        QStringList m_userAttributesOrder;
};

}

#endif // FILE_INCLUDED

