/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: extractionqueue.h 2791 2013-05-06 11:24:28Z michael $
 */

#ifndef EXTRACTIONQUEUE_INCLUDED
#define EXTRACTIONQUEUE_INCLUDED

#include "../shared/misc_defines.h"

#include <QObject>
#include <QString>
#include <QQueue>
#include <QFile>
#include <QHash>

#include "inflater.h"
#include "../shared/stopwatch.h"
#include "../shared/logsource.h"

namespace vhist
{

class VhistParser;
struct FileTag;

struct VHIST_EXTERN Extractee
{
    int id;
    int start;
    int numBytes;
    QString md5;
    QString cmd5;
    QString fullpath;
};


/**
 * A queue datastructure which extracts embedded files from a VHIST file to the
 * harddisk.  Files can be enqueued at the end of the queue and the files are
 * extracted in the order in which they are inserted into the queue.
 *
 * Each call to the process() method extracts only a limited amount of data.
 * This way, the application is still responsive in case that large files are
 * extracted.
 *
 * Information about the progress of the extraction is published via the
 * signals of the class.
 *
 * @todo More error checking
 */
class VHIST_EXTERN ExtractionQueue : public LogSource
{
    Q_OBJECT
    public: // construction, destruction, assignment
        ExtractionQueue(VhistParser* parser = NULL);
        virtual ~ExtractionQueue();
        void reset(VhistParser* parser);
        void addExtractees(QHash<int, QString> files);
    
    public slots:
        void cancelExtraction();
    
    public:
        bool process();
    
    signals:
        void fileProcessed(int id, bool valid);
        void percentFinishedChanged(int percent, const QString filename);
    
    private:
        void reinit(VhistParser* parser);
        void nextEntry();
        Extractee extracteeFromFile(const FileTag& tag);
        bool handleError(const QString& error, const Extractee& e);
    
    private: // member variables
        VhistParser* m_parser;
        Inflater m_inflater;
        StopWatch m_stopwatch;
        QQueue<Extractee> m_extractees;
        QFile m_currentFile;
        // the number of processed bytes of the current extractee
        int m_bytesHandled;
        // the total number of bytes to handle since the last reset
        qint64 m_totalBytes;
        // the total number of bytes handled since the last reset
        qint64 m_totalHandled;
};


} // namespace

#endif // EXTRACTIONQUEUE_INCLUDED

