/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: summary.h 2792 2013-05-06 11:25:27Z michael $
 */

#ifndef SUMMARY_INCLUDED
#define SUMMARY_INCLUDED

#include "../shared/misc_defines.h"

#include <QString>
#include <QStringList>
#include <QList>
#include <QHash>

namespace vhist
{

typedef QHash<QString, QString> UserAttributes;
typedef QHash<QString, QString> UserAttributeUnits;
typedef QStringList UserAttributeOrder;

/**
 * The summary of one file which appears in a workflowstep.
 */
struct VHIST_EXTERN FileSummary
{
    FileSummary();
    
    /// The id of an embedded file tag if the file is associated with
    /// an embedded file. If the summary is not identified with an
    /// embedded file (yet), \a tagId has the value -1.
    int tagID;
    /// The name of the file without a path.
    QString filename;
    /// The path of the file without the filename. It may be relative or
    /// absolute.
    QString filepath;
    /// The type of the file. This type description's convention is similar to
    /// MIME types.
    QString filetype;
    /// The description of the file.
    QString description;
    /// An additional comment refering to the file.
    QString comment;
    /// The date of the last modification of the file.
    QString lastModified;
    /// \c true if an MD5 sum has been generated for this file.
    bool generateMD5;
    /// \c true if the file has been embedded inside the VHIST file.
    bool embedded;
    /// 
    bool preview;
    /// 
    bool previewWS;
    /// 
    bool thumbnail;
    /// 
    bool thumbnailOnly;
    /// The size of the uncompressed file.
    int uncompressedSize;
    /// The MD5 sum of the uncompressed file.
    QString uncompressedMD5;
    /// The compression method used to compress the file. May be "flate" if
    /// the file is flate compressed, "none" if the file is uncompressed or
    /// an empty string if the file has not been embedded.
    QString compressedMethod;
    /// The size of the compressed file.
    int compressedSize;
    /// The MD5 sum of the compressed file.
    QString compressedMD5;
    /// A dictionary of user defined key-value pairs.
    UserAttributes userAttrs;
    /// A dictionary, which associates user attributes with units (mm, kg, etc)
    UserAttributeUnits userAttrUnits;
    /// A list which stores the order in which user attributes were given by
    /// the user.
    UserAttributeOrder userAttrOrder;
    /// The purpose of the file within the workflowstep. May be "InFile",
    /// "OutFile" or "Undefined".
    QString filePurpose;
    /// \c true if the summary is associated with a embedded file tag, which
    /// could not be matched with an entry inside one summaryfile.
    bool unmatchedTag;
};


/**
 * The summary of a workflowstep.
 */
struct VHIST_EXTERN Workflowstep
{
    Workflowstep();
    
    /// The id of a section tag which has been associated with the summary
    /// entry. \a tagID has a value of -1 if the summary entry has not
    /// been matched with any section tag.
    int tagID;
    /// The name of the vhist root file used when appending this workflowstep.
    QString vhistRootFile;
    /// The date and time at which the workflowstep has been created.
    QString timestamp;
    /// The title of the workflowstep.
    QString title;
    /// The description of the workflowstep.
    QString description;
    /// An additional comment refering to the workflowstep.
    QString comment;
    /// The command which was executed during this workflowstep.
    QString command;
    /// The tool name and version used during this workflowstep.
    QString tool;
    /// The path of the tool used during this workflowstep.
    QString toolPath;
    /// The name of the host system on which the workflowstep has been
    /// performed.
    QString host;
    /// The name of the user who executed the workflowstep.
    QString user;
    /// The name and version of the tool that created this workflowstep.
    QString creator;
    /// The title of the VHIST file shown in the PDF document properties.
    QString pdfTitle;
    /// The author of the VHIST file shown in the PDF document properties.
    QString pdfAuthor;
    /// The subject of the VHIST file shown in the PDF document properties.
    QString pdfSubject;
    /// A list of keyword shown in the PDF document properties.
    QString pdfKeywords;
    /// The producer of the VHIST file shown in the PDF document properties.
    QString pdfProducer;
    /// The creator of the VHIST file shown in the PDF document properties.
    QString pdfCreator;
    /// The list of files which are part of the workflowstep.
    QList<FileSummary> files;
    /// A list of user defined key-value pairs.
    UserAttributes userAttrs;
    /// A dictionary, which associates user attributes with units (mm, kg, etc)
    UserAttributeUnits userAttrUnits;
    /// A list which stores the order in which user attributes were given by
    /// the user.
    UserAttributeOrder userAttrOrder;
};


/**
 * The description of one summary file.
 */
struct VHIST_EXTERN Summary
{
    Summary();
    
    /// The workflowstep described in the summary file
    Workflowstep step;
    /// The md5 sum of the summary file.
    QString md5;
    /// \c true if the summary file did not contain errors.
    bool summaryOK;
};


/**
 * @typedef Summaries
 * A list of summary files.
 */
typedef QList<Summary> Summaries;


} // namespace

#endif // SUMMARY_INCLUDED

