/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: tags.h 2792 2013-05-06 11:25:27Z michael $
 */

#ifndef VHISTTAGS_INCLUDED
#define VHISTTAGS_INCLUDED

#include "../shared/misc_defines.h"

#include <QString>
#include <QByteArray>

namespace vhist
{

/**
 * A dataclass storing all entries of the tags of one embedded files.
 * Additionally, the position of the BEGIN-tag within the VHIST file
 * and the length of the tag are stored.
 */
struct VHIST_EXTERN FileTag
{
    FileTag();
    /// The id of the tag. Each filetag inside a VHIST file has a unique id.
    int id;
    /// The position of the start of the tag within the VHIST file in bytes.
    int position;
    /// The length of the tag in bytes.
    int taglen;
    /// The type of the file in a MIME-like format.
    QString filetype;
    /// The name of the file (without a path).
    QString filename;
    /// The description of the file.
    QString description;
    /// An addition comment refering to the file.
    QString comment;
    /// The type of compression used by the embedded file. Valid values are
    /// \c "none" and \c "flate".
    QString compression;
    /// The size of the uncompressed file in bytes.
    int filesize;
    /// The size of the compressed file in bytes. If the file has not been
    /// compressed, this member has a value of -1.
    int cfilesize;
    /// The size of the datablock between the \c VHIST_EMBEDDEDFILE_BEGIN
    /// and \c VHIST_EMBEDDEDFILE_END tag.
    int blocksize;
    /// The number of bytes between the end of the \c VHIST_EMBEDDEDFILE_BEGIN
    /// tag and the beginning of the embedded file.
    int offset;
    /// The md5 sum of the uncompressed file.
    QString md5file;
    /// The md5 sum of the compressed file. If the file has not been
    /// compressed, this member holds an empty string as value.
    QString md5cfile;
};


/**
 * A dataclass storing all entries of a section tag. The position of the tag
 * within the VHIST file and the length of the tag are stored, as well.
 */
struct VHIST_EXTERN SectionTag
{
    SectionTag();
    /// The id of the section. Each section has a unique id.
    int id;
    /// The position of the start of the tag within the VHIST file in bytes.
    int position;
    /// The length of the tag in bytes.
    int taglen;
    /// The version of VHIST used to create the tag.
    QString version;
    /// The name and version number of the tool which created this section.
    QString creator;
    /// The title of the section
    QString title;
    /// The number of bytes left of the start of the tag which still belong
    /// to the section.
    /// @note Section tags are usually located at the end of the section.
    int left;
    /// The size of the section in bytes.
    int size;
    /// The index of the section.
    /// @note Section indices are created in increasing order, section ids
    /// in decreasing order.
    int index;
    /// The md5 sum of the section.
    QString md5section;
    /// The md5 sum of the previous section.
    QString previousmd5;
    /// The number of bytes between the starts of the previous section tag
    /// and this section.
    int previousmarker;
    /// The section tag with the md5 sum replaced by 32 '0' (zeros). This
    /// tag is used when validating the section.
    QByteArray validationTag;
};


/**
 * A list of file tags.
 */
typedef QList<FileTag> FileTagList;

/**
 * A list of section tags.
 */
typedef QList<SectionTag> SectionTagList;

/**
 * Creates a string which can be used as a hashvalue inside a hashtable.
 */
QString VHIST_EXTERN makeFileHash(const FileTag& f, int sectionID);

} // namespace

#endif // VHISTTAGS_INCLUDED

