/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: vhistparser.h 2792 2013-05-06 11:25:27Z michael $
 */

#ifndef VHISTPARSER_INCLUDED
#define VHISTPARSER_INCLUDED

#include "../shared/misc_defines.h"

#include "tags.h"

#include "fileiterator.h"
//#include "../shared/maybe.tpl"
#include "../shared/logsource.h"

#include <QObject>
#include <QPair>
#include <QSharedPointer>
#include <QRegExp>

class QString;

namespace vhist
{


/**
 * The mode which is used by a VHIST parser to parse a file.
 */
enum ParseMode
{
    MODE_CORRECT = 0,
    MODE_ROBUST = 1
};

typedef QSharedPointer<SectionTag> MaybeSectionTag;
typedef QSharedPointer<FileTag> MaybeFileTag;
typedef QSharedPointer<QByteArray> MaybeBytes;


// <compressed sum, uncompressed sum>
typedef QPair<QString, QString> MD5Sums;


void VHIST_EXTERN printSectionTag(const SectionTag& section);
void VHIST_EXTERN printFileTag(const FileTag& file);


/**
 * @example parse_example1.cpp
 *
 * This example shows how to open a VHIST file using a vhist::VhistParser.
 */

/**
 * A parser for VHIST files. The parser can read all embedded files
 * and sections in a time and memory efficient way, extract embedded files
 * from the VHIST file and calculate checksums for them. The progress of the
 * parsing and occuring errors are published via the log signal.
 */
class VHIST_EXTERN VhistParser : public LogSource
{
    Q_OBJECT
    
    public: // construction, destruction and assignment
        VhistParser(ParseMode mode = MODE_ROBUST, QObject *pLogTarget = NULL);
        virtual ~VhistParser();
    
    public:
        bool open(const QString& filename);
        
        SectionTagList getSections();
        int getNumSections();
        
        FileTagList getFiles();
        int getNumFiles();
        FileTagList getFilesInSection(int sectionId);
        //FileTagList getFilesInNoSection();
        FileTagList getSectionlessFiles();
        
        MaybeSectionTag getSection(int id);
        MaybeFileTag getFile(int id);
        
        MaybeBytes getFileContent(int id);
        bool writeFileToDisk(int id, const QString& filename);
        MD5Sums calcChecksumsForFile(int id);
        
        QByteArray getBytes(int startPos, int numBytes);
    
    private:
        bool headerFound();
        SectionTagList readSections();
        FileTagList findFiles();
        
        QByteArray makeValidationTag(const QString& md5ToReplace);
        MaybeSectionTag findLastSection();
        MaybeSectionTag findPrevSection(const SectionTag& section);
        MaybeSectionTag parseSection(const QString& line);
        MaybeFileTag parseFile(const QString& file);
        
        bool fileInSection(const FileTag& file, const SectionTag& section);
        
        MaybeBytes getCompressedFile(int id);
        QByteArray decompressBytes(QByteArray& data,quint32 expectedSize);
    
    private:
        FileIterator m_iter;
        SectionTagList m_sections;
        FileTagList m_files;
        ParseMode m_mode;

    private:
        QRegExp m_fileRegex;
        QRegExp m_sectionRegex;
};


} // namespace

#endif // VHISTPARSER_INCLUDED

