/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: argfile.cpp 2179 2011-06-24 09:03:24Z ahuesgen $
 */

#include "../../include/arglib/argfile.h"
#include "../../include/arglib/argfilehelper.h"
#include "../../include/arglib/argfilereader.h"
#include "../../include/arglib/quote.h"

#include <QtDebug>

using namespace vhist;


/**
 * Creates a new ArgFile object.
 */
ArgFile::ArgFile() : m_verbosity(ArgFile::NORMAL), m_rootIsOptional(false),
        m_pretend(false), m_debugMode(false), m_showHelp(false),
        m_showVersion(false)
{
}


/**
 * Creates a new ArgFile object. The name of one output VHIST file and
 * the VHIST rootfile can be specified.
 *
 * @param[in] outVhistfile The filename of the VHIST file, which is generated
 * when executing this argfile.
 * @param[in] rootfile The filename of the VHIST rootfile.
 */
ArgFile::ArgFile(const QString& outVhistfile, const QString& rootfile) :
        m_verbosity(ArgFile::NORMAL), m_rootIsOptional(false), m_pretend(false),
        m_debugMode(false), m_showHelp(false), m_showVersion(false)
{
    m_outVhistfiles.append(outVhistfile);
    m_rootfile = rootfile;
}


/**
 * Copyconstructor of ArgFile.
 */
ArgFile::ArgFile(const ArgFile& rhs) :
        m_userAttributes(rhs.m_userAttributes),
        m_userAttributesOrder(rhs.m_userAttributesOrder),
        m_files(rhs.m_files),
        m_outVhistfiles(rhs.m_outVhistfiles),
        m_verbosity(rhs.m_verbosity), m_rootIsOptional(rhs.m_rootIsOptional),
        m_pretend(rhs.m_pretend), m_debugMode(rhs.m_debugMode),
        m_showHelp(rhs.m_showHelp), m_showVersion(rhs.m_showVersion),
        m_firstpage(rhs.m_firstpage),
        m_readme(rhs.m_readme),
        m_rootfile(rhs.m_rootfile),
        m_title(rhs.m_title),
        m_description(rhs.m_description),
        m_comment(rhs.m_comment),
        m_tool(rhs.m_tool),
        m_toolpath(rhs.m_toolpath),
        m_host(rhs.m_host),
        m_user(rhs.m_user),
        m_command(rhs.m_command),
        m_pdfTitle(rhs.m_pdfTitle),
        m_pdfProducer(rhs.m_pdfProducer),
        m_pdfCreator(rhs.m_pdfCreator),
        m_pdfAuthor(rhs.m_pdfAuthor),
        m_pdfKeywords(rhs.m_pdfKeywords),
        m_pdfSubject(rhs.m_pdfSubject)
{
}


/**
 * Destroys the ArgFile object.
 */
ArgFile::~ArgFile()
{
}


/**
 * Activates or deactivates the pretend mode of \e vhistadd.
 *
 * @param[in] pretend If \c true, the pretend mode of \e vhistadd is
 * activated when executing this argfile.
 */
void ArgFile::setPretend(bool pretend)
{
    m_pretend = pretend;
}


/**
 * Returns whether the pretend mode of \e vhistadd is activated or not.
 *
 * @return \c true if the pretend mode is activated, \c false otherwise.
 */
bool ArgFile::pretend() const
{
    return m_pretend;
}


/**
 * Activates or deactivates the debug mode of \e vhistadd.
 *
 * @param[in] debug If \c true, the debug mode of \e vhistadd is
 * activated when executing this argfile.
 */
void ArgFile::setDebugMode(bool debug)
{
    m_debugMode = debug;
}


/**
 * Returns whether the debug mode of \e vhistadd is activated or not.
 *
 * @return \c true if the debug mode is activated, \c false otherwise.
 */
bool ArgFile::debugMode() const
{
    return m_debugMode;
}


/**
 * Sets if \e vhistadd's help is shown when executing \e vhistadd with
 * this argfile.
 *
 * @param[in] showHelp If \c true, the help of \e vhistadd is shown.
 */
void ArgFile::setShowHelp(bool showHelp)
{
    m_showHelp = showHelp;
}


/**
 * Returns whether the help of \e vhistadd is shown when executing this
 * argfile.
 *
 * @return \c true if the help of \e vhistadd is shown.
 */
bool ArgFile::showHelp() const
{
    return m_showHelp;
}


/**
 * Sets if \e vhistadd's version information is shown when executing
 * \e vhistadd with this argfile.
 *
 * @param[in] showVersion If \c true, the version of \e vhistadd is shown.
 */
void ArgFile::setShowVersion(bool showVersion)
{
    m_showVersion = showVersion;
}


/**
 * Returns whether the version of \e vhistadd is shown when executing this
 * argfile.
 *
 * @return \c true if the version of \e vhistadd is shown.
 */
bool ArgFile::showVersion() const
{
    return m_showVersion;
}


/**
 * Sets the verbosity of \e vhistadd.
 *
 * @param[in] verbosity The verbosity of \e vhistadd.
 */
void ArgFile::setVerbosity(ArgFile::Verbosity verbosity)
{
    m_verbosity = verbosity;
}


/**
 * Returns the verbosity of \e vhistadd.
 *
 * @return The verbosity of \e vhistadd.
 */
ArgFile::Verbosity ArgFile::verbosity() const
{
    return m_verbosity;
}


/**
 * Sets the VHIST rootfile used in the workflowstep specified by
 * this argfile.
 * 
 * @param[in] rootfile The VHIST rootfle used by this argfile.
 * 
 * @note If an empty string is passed as rootfile, no rootfile will be
 * specified in the argfile.
 */
void ArgFile::setRootfile(const QString& rootfile)
{
    m_rootfile = rootfile;
}


/**
 * Returns the VHIST rootfile used in the workflowstep specified by
 * this argfile.
 * 
 * @return The VHIST rootfile used in this argfile.
 */
QString ArgFile::rootfile() const
{
    return m_rootfile;
}


/**
 * Sets whether the VHIST rootfile specified in this argfile is
 * optional or not.
 * 
 * @param[in] optional If \c true, the VHIST rootfile is optional,
 * otherwise, it is mandatory.
 */
void ArgFile::setRootIsOptional(bool optional)
{
    m_rootIsOptional = optional;
}


/**
 * Returns whether the VHIST rootfile is optional or not.
 * 
 * @return \c true if the VHIST rootfile is optional, otherwise
 * \c false.
 */
bool ArgFile::rootIsOptional() const
{
    return m_rootIsOptional;
}


/**
 * Sets the title of the workflowstep described in this argfile.
 *
 * @param[in] title The title of the workflowstep.
 */
void ArgFile::setTitle(const QString& title)
{
    m_title = title;
}


/**
 * Returns the title of the workflowstep described in this argfile.
 * 
 * @return The title of the workflowstep.
 */
QString ArgFile::title() const
{
    return m_title;
}


/**
 * Sets the description of the workflowstep described in this argfile.
 *
 * @param[in] description The description of the workflowstep.
 */
void ArgFile::setDescription(const QString& description)
{
    m_description = description;
}


/**
 * Returns the description of the workflowstep described in this argfile.
 * 
 * @return The description of the workflowstep.
 */
QString ArgFile::description() const
{
    return m_description;
}


/**
 * Sets the comment of the workflowstep described in this argfile.
 *
 * @param[in] comment The comment of the workflowstep.
 */
void ArgFile::setComment(const QString& comment)
{
    m_comment = comment;
}


/**
 * Returns the comment of the workflowstep described in this argfile.
 * 
 * @return The comment of the workflowstep.
 */
QString ArgFile::comment() const
{
    return m_comment;
}


/**
 * Sets the tool name and version of the workflowstep described in this
 * argfile.
 *
 * @param[in] tool The tool name and version of the workflowstep.
 */
void ArgFile::setTool(const QString& tool)
{
    m_tool = tool;
}


/**
 * Returns the tool name and version of the workflowstep described in this
 * argfile.
 * 
 * @return The tool name and version of the workflowstep.
 */
QString ArgFile::tool() const
{
    return m_tool;
}


/**
 * Sets the path to the tool's binary of the workflowstep described
 * in this argfile.
 *
 * @param[in] toolpath The toolpath of the workflowstep.
 */
void ArgFile::setToolpath(const QString& toolpath)
{
    m_toolpath = toolpath;
}


/**
 * Returns the path to the tool's binary of the workflowstep described
 * in this argfile.
 * 
 * @return The path to the tool's binary of the workflowstep.
 */
QString ArgFile::toolpath() const
{
    return m_toolpath;
}


/**
 * Sets the host of the workflowstep described in this argfile.
 *
 * @param[in] host The host of the workflowstep.
 */
void ArgFile::setHost(const QString& host)
{
    m_host = host;
}


/**
 * Returns the host's system name of the workflowstep described in this
 * argfile.
 * 
 * @return The host's system name of the workflowstep.
 */
QString ArgFile::host() const
{
    return m_host;
}


/**
 * Sets the user of the workflowstep described in this argfile.
 *
 * @param[in] user The user of the workflowstep.
 */
void ArgFile::setUser(const QString& user)
{
    m_user = user;
}


/**
 * Returns the user of the workflowstep described in this argfile.
 * 
 * @return The user of the workflowstep.
 */
QString ArgFile::user() const
{
    return m_user;
}


/**
 * Sets the command of the workflowstep described in this argfile.
 * 
 * @param[in] command The command of the workflowstep.
 */
void ArgFile::setCommand(const QString& command)
{
    m_command = command;
}


/**
 * Returns the command of the workflowstep described in this argfile.
 * 
 * @return The command of the workflowstep.
 */
QString ArgFile::command() const
{
    return m_command;
}


/**
 * Sets the path to the readme embedded at the beginning of the PDF
 * document generated by this argfile.
 * 
 * @param[in] readmeFile The filename of the readme file.
 */
void ArgFile::setReadme(const QString& readmeFile)
{
    m_readme = readmeFile;
}


/**
 * Returns the path to the readme embedded at the beginning of the
 * PDF document generated by this argfile.
 * 
 * @return The path to the readme embedded at the beginning of the
 * PDF document.
 */
QString ArgFile::readme() const
{
    return m_readme;
}


/**
 * Sets the path to the template for the first page of the PDF document
 * generated by this argfile.
 * 
 * @param[in] firstPageFile The filename of the firstpage template.
 */
void ArgFile::setFirstpage(const QString& firstPageFile)
{
    m_firstpage = firstPageFile;
}

/**
 * Returns the path to the template for the first page of the PDF
 * document generated by this argfile.
 *
 * @return The path to the template for the first page of the PDF
 * document.
 */
QString ArgFile::firstpage() const
{
    return m_firstpage;
}


/**
 * Sets the title of the PDF document generated with this argfile.
 *
 * @param[in] pdfTitle The title of the PDF document.
 */
void ArgFile::setPdfTitle(const QString& pdfTitle)
{
    m_pdfTitle = pdfTitle;
}


/**
 * Returns the title of the PDF document generated by this argfile.
 *
 * @return The title of the PDF document generated by this argfile.
 */
QString ArgFile::pdfTitle() const
{
    return m_pdfTitle;
}


/**
 * Sets the producer of the PDF document generated with this argfile.
 *
 * @param[in] pdfProducer The producer of the PDF document.
 */
void ArgFile::setPdfProducer(const QString& pdfProducer)
{
    m_pdfProducer = pdfProducer;
}


/**
 * Returns the producer of the PDF document generated by this argfile.
 *
 * @return The producer of the PDF document generated by this argfile.
 */
QString ArgFile::pdfProducer() const
{
    return m_pdfProducer;
}


/**
 * Sets the creator of the PDF document generated with this argfile.
 *
 * @param[in] pdfCreator The creator of the PDF document.
 */
void ArgFile::setPdfCreator(const QString& pdfCreator)
{
    m_pdfCreator = pdfCreator;
}


/**
 * Returns the creator of the PDF document generated by this argfile.
 *
 * @return The creator of the PDF document generated by this argfile.
 */
QString ArgFile::pdfCreator() const
{
    return m_pdfCreator;
}


/**
 * Sets the author of the PDF document generated with this argfile.
 *
 * @param[in] pdfAuthor The author of the PDF document.
 */
void ArgFile::setPdfAuthor(const QString& pdfAuthor)
{
    m_pdfAuthor = pdfAuthor;
}


/**
 * Returns the author of the PDF document generated by this argfile.
 *
 * @return The author of the PDF document generated by this argfile.
 */
QString ArgFile::pdfAuthor() const
{
    return m_pdfAuthor;
}


/**
 * Sets the keywords of the PDF document generated with this argfile.
 *
 * @param[in] pdfKeywords The keywords of the PDF document.
 */
void ArgFile::setPdfKeywords(const QString& pdfKeywords)
{
    m_pdfKeywords = pdfKeywords;
}


/**
 * Returns the keywords of the PDF document generated by this argfile.
 *
 * @return The keywords of the PDF document generated by this argfile.
 */
QString ArgFile::pdfKeywords() const
{
    return m_pdfKeywords;
}


/**
 * Sets the subject of the PDF document generated with this argfile.
 *
 * @param[in] pdfSubject The subject of the PDF document.
 */
void ArgFile::setPdfSubject(const QString& pdfSubject)
{
    m_pdfSubject = pdfSubject;
}


/**
 * Returns the subject of the PDF document generated by this argfile.
 *
 * @return The subject of the PDF document generated by this argfile.
 */
QString ArgFile::pdfSubject() const
{
    return m_pdfSubject;
}

/**
 * sets variables back to the default values
 *
 */
void ArgFile::clearVariables()
{
	m_verbosity=ArgFile::NORMAL;
	m_rootIsOptional=false;
	m_pretend=false;
	m_debugMode=false;
	m_showHelp=false;
	m_showVersion=false;

	m_firstpage.clear();
	m_readme.clear();
    
	m_title.clear();
	m_description.clear();
	m_comment.clear();
	m_tool.clear();
	m_toolpath.clear();
	m_host.clear();
	m_user.clear();
	m_command.clear();
    
	m_pdfTitle.clear();
	m_pdfProducer.clear();
	m_pdfCreator.clear();
	m_pdfAuthor.clear();
	m_pdfKeywords.clear();
	m_pdfSubject.clear();
}


/**
 * sets vhist root file back to the default value
 *
 */
void ArgFile::clearRootVhistfiles()
{
	m_rootfile.clear();
}

/**
 * Returns a pointer to the last file in the list of input/output files.
 * If the list is empty, NULL is returned. The pointer can be used to
 * modify the file.
 *
 * @return A pointer to the last input/output file or NULL if the list
 * is empty.
 */
WSFile* ArgFile::lastFile()
{
    if (m_files.count() > 0)
        return &m_files[m_files.count() - 1];
    else
        return NULL;
}


/**
 * Appends a file to the end of the list of input/output files.
 * A pointer to the added file is returned which can be used to
 * modify the file after it's inclusion.
 *
 * @param[in] file The input/output file, which should be added
 * to the list of input/output files.
 *
 * @return A pointer to the appended file. This pointer can be used
 * to modify the WSFile object after it's inclusion.
 *
 * @note Since a copy of \a file is appended to the list, modifying
 * the original WSFile object has no effect on the object in the list.
 */
WSFile* ArgFile::addFile(const WSFile& file)
{
    m_files.append(file);
    return lastFile();
}


/**
 * Returns the number of input/output files in the workflowstep
 * described in this argfile.
 *
 * @return The number of input/output files.
 */
int ArgFile::numFiles() const
{
    return m_files.count();
}


/**
 * Returns a pointer to the i-th file in the list of
 * input/output files. If the given \a index is not valid,
 * \c NULL is returned.
 * 
 * @param[in] index The index of the file, which should be returned.
 * Indices are 0-based.
 * 
 * @return A pointer to the file or \c NULl if the index is not valid.
 * 
 * @note The pointer returned is only valid as long as the file in inside
 * the filelist.
 */
WSFile* ArgFile::getFile(int index)
{
    if (m_files.count() > index)
        return &m_files[index];
    else
        return NULL;
}


/**
 * Returns a copy of the i-th file in the list of input/output files.
 * If the given \a index is not valid, an object generated by
 * the default constructor is returned. This function can be used to
 * get input/output files of constant WSFile objects.
 * 
 * @param[in] index The index of the file, which should be returned.
 * Indices are 0-based.
 * 
 * @return A copy to the file.
 */
WSFile ArgFile::getConstFile(int index) const
{
    if (m_files.count() > index)
        return m_files[index];
    else
        return WSFile();
}


/**
 * Removes all files from the list of input/output files.
 */
void ArgFile::clearFilelist()
{
    m_files.clear();
}


/**
 * Removes the file at the i-th position inside the queue.
 * If \a index is invalid, no element is removes.
 *
 * @param[in] index The index of the file to remove.
 */
void ArgFile::removeFile(int index)
{
    if (index >= 0 && index < m_files.size())
        m_files.removeAt(index);
}


/**
 * Adds a user attribute to the workflowstep. If a user attribute with
 * the same key does already exist, this attribute is replaced by the
 * new attribute.
 * 
 * @param[in] key The key of the attribute.
 * @param[in] value The value of the attribute.
 */
void ArgFile::addUserAttribute(const QString& key, const QString& value)
{
    m_userAttributesOrder.append(key);
    m_userAttributes.insert(key, value);
}


/**
 * Removes all user attributes.
 */
void ArgFile::clearUserAttributes()
{
    m_userAttributes.clear();
    m_userAttributesOrder.clear();
}


/**
 * Returns a list of the keys of all user attributes in the
 * workflowstep described by this argfile. This list can be
 * used to iterate over all attributes.
 * 
 * @return A list of all user attribute keys.
 */
QStringList ArgFile::userAttributeKeys() const
{
    return m_userAttributesOrder;
}


/**
 * Returns the value of the user attribute attribute with
 * key \a key. If no user attribute with this key exists,
 * an empty string is returned.
 * 
 * @param[in] key The key of the user attribute to return.
 * 
 * @return The value belonging to the user attribute with the
 * given key.
 */
QString ArgFile::userAttribute(const QString& key) const
{
    return m_userAttributes.value(key, "");
}


/**
 * Adds a file to the list of output vhistfiles.
 * 
 * @param[in] filename The filename of the output vhistfile.
 */
void ArgFile::addOutputVhistfile(const QString& filename)
{
    m_outVhistfiles.append(filename);
}


/**
 * Removes all output vhistfiles.
 */
void ArgFile::clearOutputVhistfiles()
{
    m_outVhistfiles.clear();
}


/**
 * Returns the number of output vhistfiles.
 * 
 * @return The number of output vhistfiles.
 */
int ArgFile::numOutputVhistfiles() const
{
    return m_outVhistfiles.count();
}


/**
 * Returns the i-th output vhistfile.
 * 
 * @param[in] index The index of the output vhistfile to return.
 * \a index is 0-based. If the \a index is not valid, an empty string
 * is returned.
 * 
 * @return The requested output vhistfile.
 */
QString ArgFile::outputVhistfile(int index) const
{
    if (m_outVhistfiles.count() > index)
        return m_outVhistfiles[index];
    else
        return "";
}


/**
 * Generates a string containing the content of an argfile representing
 * this ArgFile object.
 * 
 * @param supportVariables If \c true, strings of the form \c $NAME inside
 * the ArgFile object are regarded as variables and are not escaped inside
 * the generated argfile. To insert a literal \c $ sign inside the argfile,
 * the \c $ sign must be doubled. If \c false, all \c $ sign is inserted as
 * a literal \c $ sign into the argfile.
 * 
 * @return The generated argfile.
 */
QString ArgFile::toCommandline(bool supportVariables) const
{
    QStringList list;
    
    if (m_verbosity == ArgFile::QUIET)
        list << "--quiet";
    else if (m_verbosity == ArgFile::VERBOSE)
        list << "--verbose";
    
    if (m_pretend)
        list << "--pretend";
    
    if (m_debugMode)
        list << "--debug";
    
    if (m_showHelp)
        list << "--help";
    
    if (m_showVersion)
        list << "--version";
    
    if (m_firstpage != "")
        list << "--firstpage" << m_firstpage;
    
    if (m_readme != "")
        list << "--readme" << m_readme;
    
    for (int i = 0; i < m_outVhistfiles.size(); i++)
        list << "-O" << m_outVhistfiles[i];
    
    if (m_rootfile != "")
        list << (m_rootIsOptional ? "-J" : "-I") << m_rootfile;
    
    // section attributes
    appendAttributeToList(m_title, "-s", "title", list);
    appendAttributeToList(m_description, "-s", "description", list);
    appendAttributeToList(m_comment, "-s", "comment", list);
    appendAttributeToList(m_tool, "-s", "tool", list);
    appendAttributeToList(m_toolpath, "-s", "toolpath", list);
    appendAttributeToList(m_host, "-s", "host", list);
    appendAttributeToList(m_user, "-s", "user", list);
    appendAttributeToList(m_command, "-s", "command", list);
    
    // document attributes
    appendAttributeToList(m_pdfTitle, "-d", "title", list);
    appendAttributeToList(m_pdfProducer, "-d", "producer", list);
    appendAttributeToList(m_pdfCreator, "-d", "creator", list);
    appendAttributeToList(m_pdfAuthor, "-d", "author", list);
    appendAttributeToList(m_pdfKeywords, "-d", "keywords", list);
    appendAttributeToList(m_pdfSubject, "-d", "subject", list);
    
    QStringList keys = userAttributeKeys();
    for (int i = 0; i < keys.size(); i++)
    {
        QString key = keys.at(i);
        QString value = m_userAttributes[key];
        appendAttributeToList(value, "-U", key, list);
    }
    
    QString cmdline = quoteStringList(list, supportVariables).join(" ");
    
    for (int i = 0; i < m_files.size(); i++)
    {
        WSFile f= m_files.at(i);
        cmdline += QString(" ") + f.toCommandline(supportVariables);
    }
    
    return cmdline;
}

QString ArgFile::getUndoWorkflowStep() const
{
	QString szUndo;
    for (int i = 0; i < m_outVhistfiles.size(); i++)
		szUndo+=QString("<DelWorkflowStep>%1</DelWorkflowStep>")
			.arg(m_outVhistfiles[i]); 
	return szUndo;
}

