/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

Author: Michael May
$Id: wsfile.cpp 2179 2011-06-24 09:03:24Z ahuesgen $
*/

#include "../../include/arglib/wsfile.h"
#include "../../include/arglib/argfilehelper.h"
#include "../../include/arglib/quote.h"

#include <QFileInfo>
#include <QDir>

using namespace vhist;

/**
 * The default constructor of WSFile.
 */
WSFile::WSFile() : m_md5(true), m_embed(true), m_optional(false),
        m_compress(false), m_previewws(false), m_preview(false),
        m_thumbnail(false), m_thumbnailOnly(false),
        m_filePurpose(WSFile::INPUT_FILE)
{
}


/**
 * The copy constructor of WSFile.
 * 
 * @param[in] rhs The WSFile object, which should be copied.
 */
WSFile::WSFile(const WSFile& rhs) :
         m_filetype(rhs.m_filetype),
         m_filename(rhs.m_filename),
         m_filepath(rhs.m_filepath),
         m_md5File(rhs.m_md5File),
         m_md5(rhs.m_md5),
         m_embed(rhs.m_embed),
         m_optional(rhs.m_optional),
         m_compress(rhs.m_compress),
         m_previewws(rhs.m_previewws),
         m_preview(rhs.m_preview),
         m_thumbnail(rhs.m_thumbnail),
         m_thumbnailOnly(rhs.m_thumbnailOnly),
         m_filePurpose(rhs.m_filePurpose),
         m_description(rhs.m_description),
         m_comment(rhs.m_comment),
         m_userAttributes(rhs.m_userAttributes),
         m_userAttributesOrder(rhs.m_userAttributesOrder)
{
}


bool WSFile::operator==(const WSFile& rhs)
{
    return (m_filetype == rhs.m_filetype) &&
            (m_filename == rhs.m_filename) &&
            (m_filepath == rhs.m_filepath) &&
            (m_md5File == rhs.m_md5File) &&
            (m_md5 == rhs.m_md5) &&
            (m_embed == rhs.m_embed) &&
            (m_optional == rhs.m_optional) &&
            (m_compress == rhs.m_compress) &&
            (m_preview == rhs.m_preview) &&
            (m_previewws == rhs.m_previewws) &&
            (m_thumbnail == rhs.m_thumbnail) &&
            (m_thumbnailOnly == rhs.m_thumbnailOnly) &&
            (m_filePurpose == rhs.m_filePurpose) &&
            (m_comment == rhs.m_comment) &&
            (m_description == rhs.m_description) &&
            (m_userAttributes == rhs.m_userAttributes) &&
            (m_userAttributesOrder == rhs.m_userAttributesOrder);
}


bool WSFile::operator!=(const WSFile& rhs)
{
    return !(operator==(rhs));
}


/**
 * Sets the description of the file.
 *
 * @param[in] description Text describing the file.
 */
void WSFile::setDescription(const QString& description)
{
    m_description = description;
}


/**
 * Returns the current file description text.
 *
 * @return The current file description text.
 */
QString WSFile::description() const
{
    return (m_description);
}


/**
 * Sets the comment of the file.
 * 
 * @param[in] comment A comment concerning the file.
 */
void WSFile::setComment(const QString& comment)
{
    m_comment = comment;
}


/**
 * Returns the comment of the file.
 *
 * @return The comment of the file.
 */
QString WSFile::comment() const
{
    return m_comment;
}


/**
 * Sets whether the input/output file should be embedded into the
 * VHIST file.
 * 
 * @param embed \c true if the file should be embedded into the VHIST
 * file.
 */
void WSFile::setEmbed(bool embed)
{
    m_embed = embed;
}


/**
 * Returns whether the input/output file should be embedded into
 * the VHIST file.
 * 
 * @return \c true if the file should be embedded into the VHIST file,
 * \c false otherwise.
 */
bool WSFile::embed() const
{
    return m_embed;
}


/**
 * Sets the filetype associated with the file.
 * 
 * @param[in] filetype The filetype of the file.
 * 
 * @note The \a filetype attribute is a freeform string, however, it
 * is recommended to format the filetype in a MIME type like format.
 */
void WSFile::setFiletype(const QString& filetype)
{
    m_filetype = filetype;
}


/**
 * Returns the current filetype of this input/output file.
 * 
 * @return The current filetype.
 */
QString WSFile::filetype() const
{
    return (m_filetype);
}


/**
 * Sets the filename and the file path as one string.
 * 
 * @param[in] filename The path + name of the file.
 */
void WSFile::setFilenameAndPath(const QString& filename)
{
    QFileInfo f = filename;

    setFilepath(f.dir().path());
    setFilename(f.fileName());
}


/**
 * Sets whether the file is optional or not. If the file should be embedded
 * into the VHIST file and is not found by \e vhistadd, this flag determines
 * whether \e vhistadd emits an error or ignores the missing file.
 * 
 * @param[in] optional \c true if the file is optional, \c false otherwise.
 */
void WSFile::setOptional(bool optional)
{
    m_optional = optional;
}


/**
 * This functions returns whether the file is optional.
 *
 * @return \c true if the file is optional, \c false otherwhise.
 */
bool WSFile::optional() const
{
    return m_optional;
}


/**
 * Sets whether the file is an input or an output file to the workflowstep.
 * 
 * @param[in] inputFile \c true if this file is an input file,
 * \c false if it is an output file.
 */
void WSFile::setInputFile(bool inputFile)
{
    m_filePurpose = inputFile ? WSFile::INPUT_FILE : WSFile::OUTPUT_FILE;
}


/**
 * This function returns whether the file is an input or output file. 
 * 
 * @return \c true if the file is an input file, \c false if it is an
 * output file.
 */
bool WSFile::inputFile() const
{
    return (m_filePurpose == WSFile::INPUT_FILE);
}


/**
 * Returns the current filepath and filename as one string.
 *
 * @return The filepath and filename.
 */
QString WSFile::filename() const
{
    return (m_filepath + m_filename);
}


/**
 * Sets the name of the file without the path.
 * 
 * @param[in] name The filename without path.
 */
void WSFile::setFilename(const QString& name)
{
    m_filename = name;
}

/**
 * Returns the filepath without the filename.
 *
 * @return The filepath without the filename.
 */
QString WSFile::filepath() const
{
    return m_filepath;
}

/**
 * Sets the filepath without the filename.
 *
 * @param[in] path The filepath.
 */
void WSFile::setFilepath(const QString& path)
{
    m_filepath = path;
    
    /* Make sure it ends with / */
    // TODO add windows separator support
    if (!m_filepath.endsWith ('/')) {
        m_filepath = m_filepath + "/";
    }
}


/**
 * Sets whether \e vhistadd should automatically generate an MD5 checksum
 * for this file.
 * 
 * @param[in] md5 \c true if a checksum should be calculate and
 * false otherwise.
 */
void WSFile::setAutoMD5(bool md5)
{
   m_md5 = md5;    
}


/**
 * This function returns whether \e vhistadd should automatically
 * generate an MD5 checksum for the file.
 * 
 * @return \c true if a checksum should be calculated, false otherwise.
 */
bool WSFile::autoMD5() const
{
    return m_md5;
}


/**
 * Adds a new user-defined attribute to the input/output file.
 * An attribute is an arbitrary key-value pair. If another attribute
 * with the same key does already exist, it is overwritten by the new
 * attribute.
 * 
 * @param[in] key The key of the attribute.
 * @param[in] value The value of the attribute.
 */
void WSFile::addUserAttribute(const QString& key, const QString& value)
{
    m_userAttributesOrder.append(key);
    m_userAttributes.insert(key, value);
}


/**
 * Removes all user-defined attributes from the input/output file.
 */
void WSFile::clearUserAttributes()
{
    m_userAttributesOrder.clear();
    m_userAttributes.clear();
}


/**
 * Returns a list of all keys appearing in user defined attributes.
 * This list can be used to iterate over all attributes.
 * 
 * @return A list of keys.
 */
QStringList WSFile::userAttributeKeys() const
{
    return m_userAttributesOrder;
}


/**
 * Returns the value associated with the given \a key.
 * If the key does not appear in any attribute, an
 * empty string is returned.
 * 
 * @param[in] key The key of a user-defined attribute.
 */
QString WSFile::userAttribute(const QString& key) const
{
    return m_userAttributes.value(key, "");
}


/**
 * Creates a string containing the vhistadd commandline for this file.
 *
 * @return The commandline needed to describe this file in vhistadd.
 */
QString WSFile::toCommandline(bool supportVariables)
{
    QStringList list;
    
    list.append((m_filePurpose == WSFile::INPUT_FILE) ? "-i" : "-o");
    list.append(QFileInfo(QDir(m_filepath), m_filename).filePath());

    appendAttributeToList(m_filetype, "-a", "filetype", list);
    appendAttributeToList(m_description, "-a", "description", list);
    appendAttributeToList(m_comment, "-a", "comment", list);
    appendAttributeToList(m_md5File, "-a", "md5file", list);
    
    appendFlagToList(m_optional, "-f", "optional", list);
    appendFlagToList(m_compress, "-f", "compress", list);
    appendFlagToList(m_embed, "-f", "embed", list);
    appendFlagToList(m_md5, "-f", "automd5", list);
    appendFlagToList(m_previewws, "-f", "previewws", list);
    appendFlagToList(m_preview, "-f", "preview", list);
    appendFlagToList(m_thumbnail, "-f", "thumbnail", list);
    appendFlagToList(m_thumbnailOnly, "-f", "thumbnailonly", list);
    
    QStringList keys = userAttributeKeys();
    for (int i = 0; i < keys.size(); i++)
    {
        QString key = keys.at(i);
        QString value = m_userAttributes.value(key);
        appendAttributeToList(value, "-u", key, list);
    }
    
    return quoteStringList(list, supportVariables).join(" ");
}


/**
 * Returns the purpose of the file in the workflowstep. A file can be either
 * an input- or an outputfile.
 *
 * @return The purpose of the file in the workflowstep.
 */
WSFile::FilePurpose WSFile::filePurpose() const
{
    return m_filePurpose;
}


/**
 * Sets the purpose of the file in the workflowstep.
 * 
 * @param[in] purpose The purpose of the file in the workflowstep.
 */
void WSFile::setFilePurpose(const WSFile::FilePurpose& purpose)
{
    m_filePurpose = purpose;
}


/**
 * Sets an MD5 file for this input/output file. An MD5 file is a file
 * which contains an MD5 checksum in hexdigests in it's first line.
 * 
 * @param[in] file The name of the MD5 file.
 */
void WSFile::setMD5File(const QString& file)
{
    m_md5File = file;
}


/**
 * Returns the MD5 file of this input/output file.
 *
 * @return The MD5 file of this input/output file.
 */
QString WSFile::md5File() const
{
    return m_md5File;
}


/**
 * Returns whether the previewws flag is active for this file.
 * 
 * @return \c true if the previewws flag is active for this file,
 * \c false otherwise.
 */
bool WSFile::previewWS() const
{
	return m_previewws;
}


/**
 * Activates/deactivates the previewws flag for this file.
 * @param[in] previewWS \c true if the flag should be activated,
 * \c false otherwise.
 */
void WSFile::setPreviewWS(bool previewWS)
{
	m_previewws = previewWS;
}


/**
 * Returns whether the preview flag is active for this file.
 * 
 * @return \c true if the preview flag is active for this file,
 * \c false otherwise.
 */
bool WSFile::preview() const
{
	return m_preview;
}

/**
 * Activates/deactivates the preview flag for this file.
 * @param[in] preview \c true if the flag should be activated,
 * \c false otherwise.
 */
void WSFile::setPreview(bool preview)
{
	m_preview = preview;
}


/**
 * Returns whether the thumbnail flag is active for this file.
 * 
 * @return \c true if the thumbnail flag is active for this file,
 * \c false otherwise.
 */
bool WSFile::thumbnail() const
{
	return m_thumbnail;
}

/**
 * Activates/deactivates the thumbnail flag for this file.
 * @param[in] thumbnail \c true if the flag should be activated,
 * \c false otherwise.
 */
void WSFile::setThumbnail(bool thumbnail)
{
	m_thumbnail = thumbnail;
}


/**
 * Returns whether the thumbnailonly flag is active for this file.
 * 
 * @return \c true if the thumbnailonly flag is active for this file,
 * \c false otherwise.
 */
bool WSFile::thumbnailOnly() const
{
	return m_thumbnailOnly;
}


/**
 * Activates/deactivates the thumbnailonly flag for this file.
 * @param[in] thumbnailOnly \c true if the flag should be activated,
 * \c false otherwise.
 */
void WSFile::setThumbnailOnly(bool thumbnailOnly)
{
	m_thumbnailOnly = thumbnailOnly;
}


/**
 * Sets whether the input/output file should be embedded in compressed
 * or uncompressed form. If the file is not embedded anyway, this flag
 * is ignored by \e vhistadd.
 * 
 * @param[in] compress \c true if the file should be compressed, \c false
 * otherwise.
 */
void WSFile::setCompress(bool compress)
{
    m_compress = compress;
}


/**
 * Returns whether the input/output file should be embedded in compressed
 * or uncompressed form.
 * 
 * @return \c true if the file should be compressed, \c false otherwise.
 */
bool WSFile::compress() const
{
    return m_compress;
}

