/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: logsource.cpp 2187 2011-06-27 11:59:35Z ahuesgen $
 */

#include "../../include/shared/logsource.h"

#include <QtDebug>


using namespace vhist;

/**
 * Creates a new LogSource object and registers a listener
 * with it. The listener must provide a slot with signature
 * void log(const QString, const vhist::LogSource::LogType)
 * or void log(const QString). The listener will receive all
 * log messages generated by this class via the slot.
 *
 * @param[in] listener A class which should listen to the
 * log messages of the LogSource object. Can be \c NULL,
 * in which case no object is registered.
 */
LogSource::LogSource(QObject* listener)
{
    publishToLogSlot(listener);
}


/**
 * Destroys a LogSource object.
 */
LogSource::~LogSource()
{
}


/**
 * Registers a listener with the LogSource object.
 * The listener must provide a slot with signature
 * void log(const QString, const int)
 * or void log(const QString). The listener will receive all
 * log messages generated by this class via the slot.
 *
 * @param[in] listener A class which should listen to the
 * log messages of the LogSource object. Can be \c NULL,
 * in which case no object is registered.
 */
void LogSource::publishToLogSlot(QObject* listener)
{
    if (listener == NULL)
        return;
    
    // neccessary slots and signals
    const char* slot1 = SLOT(log(const QString,
            const int));
    const char* slot2 = SLOT(log(const QString));
    const char* signal1 = SIGNAL(log(const QString,
            const int));
    const char* signal2 = SIGNAL(log(const QString));
    
    if (listener->metaObject()->indexOfSlot(slot1) != -1)
        connect(this, signal1, listener, slot1);
    else
        connect(this, signal2, listener, slot2);
}


/**
 * Registers a listener with the LogSource object.
 * The listener must provide a signal with signature
 * void log(const QString, const int)
 * or void log(const QString). All log messages will be
 * forwarded to this signal.
 *
 * @param[in] listener A class which should listen to the
 * log messages of the LogSource object. Can be \c NULL,
 * in which case no object is registered.
 */
void LogSource::publishToLogSignal(QObject* listener)
{
    if (listener == NULL)
        return;
    
    // neccessary signals
    const char* signal1 = SIGNAL(log(const QString,
            const int));
    const char* signal2 = SIGNAL(log(const QString));
    
    if (listener->metaObject()->indexOfSignal(signal1) != -1)
        connect(this, signal1, listener, signal1);
    else
        connect(this, signal2, listener, signal2);
}


/**
 * @fn void vhist::LogSource::log(const QString message, const int)
 * 
 * This signal is emitted by the LogSource to publish
 * a log message. The message has a type which tells
 * whether its a simple message, a warning or an
 * error.
 * 
 * @param[in] message The log message.
 * @param[in] type The type of the log message. Valid values are all values
 * in the \c enum vhist::LogSource::LogType.
 */

