/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: extractionqueue.cpp 2458 2011-08-03 09:21:52Z ahuesgen $
 */

#include "../../include/unvhistlib/extractionqueue.h"
#include "../../include/unvhistlib/vhistparser.h"

#include <QtDebug>

using namespace vhist;

/**
 * Creates a new ExtractionQueue object. The object is identified with
 * a VhistParser object. All embedded files are taken from this parser.
 *
 * @param[in] parser The parser associated with the ExtractionQueue.
 * The \a parser must not be \c NULL.
 */
ExtractionQueue::ExtractionQueue(VhistParser* parser) : m_totalBytes(0)
{
    m_stopwatch.publishToLogSignal(this);
    reinit(parser);
}


/**
 * Destroys an ExtractionQueue object.
 */
ExtractionQueue::~ExtractionQueue()
{
}


/**
 * Resets the ExtractionQueue object and associates the queue with
 * another VhistParser object.
 *
 * @param[in] parser The new VhistParser object. Must not be \c NULL.
 */
void ExtractionQueue::reset(VhistParser* parser)
{
    reinit(parser);
    //m_stopwatch.start("validation finished");
}


void ExtractionQueue::reinit(VhistParser* parser)
{
    m_extractees.clear();
    m_parser = parser;
    m_totalBytes = 0;
    m_totalHandled = 0;
    m_bytesHandled = 0;
}


/**
 * Cancels the extraction currently running and clears the queue. If the
 * current file has not been extracted completely, the incomplete file
 * will still be in the filesystem.
 */
void ExtractionQueue::cancelExtraction()
{
    m_currentFile.close();
    reinit(m_parser);
    m_stopwatch.stop();
}


/**
 * Adds a list of files, which should be extracted, to the ExtractionQueue.
 * 
 * @param[in] filesToExtract A hashtable with file ids as keys and
 * the filenames of the files on disk as values.
 * 
 * @note Due to the nature of hashtables, the order in which the files
 * are inserted into the queue is probably not the same order in which
 * they have been inserted into the hashtable.
 */
void ExtractionQueue::addExtractees(QHash<int, QString> filesToExtract)
{
    if (m_parser != NULL)
    {
        FileTagList files = m_parser->getFiles();
        
        for (int i = 0; i < files.count(); i++)
        {
            const FileTag& f = files.at(i);
            if (f.id != -1 && filesToExtract.contains(f.id))
            {
                Extractee e = extracteeFromFile(f);
                e.fullpath = filesToExtract[f.id];
                m_extractees.enqueue(e);
                m_totalBytes += e.numBytes;
            }
        }
    }
}


void ExtractionQueue::nextEntry()
{
    if (!m_extractees.empty())
    {
        Extractee e = m_extractees.head();
        // check if file has not been processed completely and
        // correct statistics if needed
        m_totalHandled += (e.numBytes - m_bytesHandled);
        m_extractees.dequeue();
    }
    m_bytesHandled = 0;
    m_currentFile.close();
}


Extractee ExtractionQueue::extracteeFromFile(const FileTag& tag)
{
    Extractee e;
    e.id = tag.id;
    e.start = tag.position + tag.taglen + tag.offset;
    e.md5 = tag.md5file;
    
    if (tag.compression == "flate")
    {
        e.numBytes = tag.cfilesize;
        e.cmd5 = tag.md5cfile;
    }
    else
    {
        e.numBytes = tag.filesize;
        e.cmd5 = "";
    }
    return e;
}


bool ExtractionQueue::handleError(const QString& error, const Extractee& e)
{
    emit log(error, LogSource::LOG_ERROR);
    emit fileProcessed(e.id, false);
    nextEntry();
    return !m_extractees.empty();
}


/**
 * Continues extraction of the current file to disk. The method will only
 * process a limited amount of data before it returns. This way, the
 * application stays responsive, even when large files are extracted. This
 * is especially important for applications with graphical userinterfaces.
 * 
 * @return \c false if the last file has been extracted and the queue is
 * empty, now. Otherwise \c true.
 */
bool ExtractionQueue::process()
{
    QString currentName;
    if (!m_extractees.empty())
    {
        // get current extractee
        Extractee e = m_extractees.head();
        currentName = e.fullpath;
        if (m_bytesHandled == 0)
        {
            // new file => reset everything
            m_inflater.reset();
            m_currentFile.close();
            m_currentFile.setFileName(e.fullpath);
            m_currentFile.open(QIODevice::WriteOnly | QIODevice::Unbuffered);
        }
        
        // process 100000 bytes at most
        int numBytes = qMin(e.numBytes - m_bytesHandled, 100000);

        // read data from embedded file
        QByteArray bytes = m_parser->getBytes(e.start + m_bytesHandled,
                numBytes);
        if (numBytes != bytes.size())
        {
            QString err = QString("Could not read from file %1");
            return handleError(err.arg(currentName), e);
        }
        m_bytesHandled += numBytes;
        m_totalHandled += numBytes;
        
        // deflate if neccessary
        if (e.cmd5 != "")
        {
            bytes = m_inflater.inflateData(bytes);
            
            if (!m_inflater.isValid())
            {
                QString err = QString("Could not decompress data from file %1");
                return handleError(err.arg(currentName), e);
            }
        }
        
        // write data to disk
        int bytesWritten = m_currentFile.write(bytes);
        if (bytesWritten != bytes.size())
        {
            QString err = QString("Could not write to file file %1");
            return handleError(err.arg(currentName), e);
        }
        
        // finished?
        if (m_bytesHandled == e.numBytes)
        {
            //qDebug() << "file: " << e.id;
            emit fileProcessed(e.id, true);
            nextEntry();
        }
    }
    
    // calculate statistics
    int percent = static_cast<int>(100.0 * m_totalHandled / m_totalBytes);
    emit percentFinishedChanged(percent, currentName);
    
    if (m_extractees.empty())
    {
        m_stopwatch.stop();
        m_bytesHandled = 0;
        m_totalBytes = 0;
        m_totalHandled = 0;
        m_currentFile.close();
    }
    
    return !m_extractees.empty();
}



// documentation of Qt signals

/**
 * @fn void ExtractionQueue::fileProcessed(int id, bool valid)
 * 
 * This signal is emitted each time a file has been completely processed.
 * 
 * @param[in] id The id of the processed file.
 * @param[in] valid \c true if the file was valid and extraction succeeded.
 */

/**
 * @fn void ExtractionQueue::percentFinishedChanged(int percent, const QString filename)
 *
 * This signal is emitted after each call to process and informs all connected
 * slots about the percentage of work finished, yet. It also states the name
 * of the file currently processed.
 *
 * @param[in] percent The percentage of work already done.
 * @param[in] filename The name of the file which is currently extracted.
 */

