/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: inflater.cpp 2793 2013-05-06 11:26:17Z michael $
 */

#include "../../include/unvhistlib/inflater.h"

#include <QtDebug>

using namespace vhist;


/**
 * Creates a new Inflater.
 */
Inflater::Inflater()
{
    m_iChunk = 1024 * 1024 * 1;
    m_out = (unsigned char *) malloc (m_iChunk);
    init();
}


/**
 * Destroys the Inflater.
 */
Inflater::~Inflater()
{
    inflateEnd(&m_stream);
    free(m_out);
}


/**
 * Resets the Inflater. This way, a new datastream can be deflated
 * by the same Inflater object.
 */
void Inflater::reset()
{
    inflateEnd(&m_stream);
    init();
}


/**
 * Inflates some data from the datastream. You can check if the inflation
 * was successful by querying the isValid() method of the Inflater.
 *
 * @param[in] data The data which should be inflated.
 *
 * @return The inflated data. If an error occures during inflating
 * the data, an empty QByteArray is returned.
 */
QByteArray Inflater::inflateData(QByteArray data)
{
    QByteArray output;

    if (data.count() > 0)
    {
        m_stream.avail_in = data.count();
        m_stream.next_in = reinterpret_cast<unsigned char*>(data.data());
        do
        {
            m_stream.avail_out = m_iChunk;
            m_stream.next_out = m_out;
            int ret = inflate(&m_stream, Z_NO_FLUSH);
            if (ret != Z_OK && ret != Z_STREAM_END)
            {
                inflateEnd(&m_stream);
                m_valid = false;
                return QByteArray();
            }
            
            m_finished = (ret == Z_STREAM_END);
            if (m_finished){
                break;
            }

            if (m_stream.avail_out == 0)
                output.append(QByteArray(reinterpret_cast<char*>(m_out), m_iChunk));
        
        } while (m_stream.avail_out == 0);
        
        // avail_out contains the free (unused) space in out
        int rest = m_iChunk - m_stream.avail_out;
        output.append(QByteArray(reinterpret_cast<char*>(m_out), rest));
    }
    return output;
}


/**
 * Returns whether the datastream has terminated, yet. This is convenient
 * if the size of the datastream is not known.
 *
 * @return Whether the datastream has reached an end.
 */
bool Inflater::hasFinished()
{
    return m_finished;
}


/**
 * Returns \c true if the datastream has not been corrupted so far.
 *
 * @return \c true if no corruption of the datastream did occur so far.
 */
bool Inflater::isValid()
{
    return m_valid;
}




// private methods

void Inflater::init()
{
    m_valid = true;
    m_finished = false;
    
    // allocate inflate state
    m_stream.zalloc = Z_NULL;
    m_stream.zfree = Z_NULL;
    m_stream.opaque = Z_NULL;
    m_stream.avail_in = 0;
    m_stream.next_in = Z_NULL;
    int ret = inflateInit(&m_stream);
    if (ret != Z_OK)
        m_valid = false;
}
