/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: summarycreator.cpp 2458 2011-08-03 09:21:52Z ahuesgen $
 */

#include <QString>
#include <QtDebug>
#include <QPair>

#include "../../include/unvhistlib/summarycreator.h"
#include "../../include/unvhistlib/vhistparser.h"
#include "../../include/unvhistlib/summaryreader.h"
#include "../../include/unvhistlib/summary.h"

using namespace vhist;

const int FILEDIST_THRESHOLD = 15;

// a filetag plus the section id of the section, in which the file is located
typedef QPair<FileTag, int> SectionFileTag;
typedef QMultiHash<QString, SectionFileTag> TagHashTable;

// first: id of tag, second: distance
typedef QPair<int, int> FileDistance;

// function forward declarations
TagHashTable makeTagHash(VhistParser& parser);
Summaries makeSummaries(VhistParser& parser, TagHashTable filetags);
bool filesEqual(const FileSummary& summary, const FileTag& tag);
bool workflowstepsEqual(const Workflowstep&, const SectionTag&);
int distance(const FileTag& tag, const FileSummary& summary);
Summary augmentSummary(const Summary& summary, const SectionTag& section,
        TagHashTable& filetags);
Summary makeUnmatchedTagSummary(const FileTagList& unmatched);


// methods of SummaryCreator

/**
 * Creates a new SummaryCreator instance.
 */
SummaryCreator::SummaryCreator()
{
}


/**
 * Searches the given VhistParser for summary files and associates them
 * with tags inside the VHIST file. The resulting Summaries are returned.
 *
 * @param[in] parser A parser with an opened VHIST file, for which the
 * vhist::Summaries should be created.
 *
 * @return A list of Summaries
 *
 * @note If a tag inside the VHIST file could not be matched with an
 * entry in a summary, a dummy entry is generated for this file. This way,
 * no information is hidden if the VHIST file contains errors.
 */
Summaries SummaryCreator::createSummaries(VhistParser& parser)
{
    TagHashTable tags = makeTagHash(parser);
    return makeSummaries(parser, tags);
}


// helper functions (not part of SummaryCreator)

TagHashTable makeTagHash(VhistParser& parser)
{
    TagHashTable tagsHash;
    SectionTagList sections = parser.getSections();
    for (int i = 0; i < sections.count(); i++)
    {
        int sectionID = sections.at(i).id;
        FileTagList files = parser.getFilesInSection(sectionID);
        for (int j = 0; j < files.count(); j++)
        {
            FileTag filetag = files.at(j);
            QString hash = makeFileHash(filetag, sectionID);
            tagsHash.insert(hash, SectionFileTag(filetag, sectionID));
        }
    }
    return tagsHash;
}


Summaries makeSummaries(VhistParser& parser, TagHashTable filetags)
{
    Summaries result;
    
    SectionTagList sections = parser.getSections();
    
    for (int i = sections.count()-1; i >= 0; i--)
    {
        SectionTag s = sections.at(i);
        FileTagList files = parser.getFilesInSection(s.id);
        
        bool summaryCreated = false;
        if (files.count() > 0)
        {
            FileTag summaryFile = files.last();
            if (summaryFile.filename == "ws_summary.xml")
            {
                MaybeBytes maybe = parser.getFileContent(summaryFile.id);
                if (maybe != NULL)
                {
                    QByteArray data = *maybe;
                    SummaryReader reader;
                    Summary summary = reader.readSummary(
                            QString::fromUtf8(data));
                    summary = augmentSummary(summary, s, filetags);
                    result.append(summary);
                    summaryCreated = true;
                }
                else
                    qDebug() << "could not read summary" << summaryFile.id;
            }
        }
        
        // summary file not found or broken
        if (!summaryCreated)
        {
            Summary summary = makeSummaryFromTags(s, files);
            result.append(summary);
        }
        
    }
    
    Summary unmatched = makeUnmatchedTagSummary(parser.getSectionlessFiles());
    if (unmatched.step.files.count() > 0)
        result.append(unmatched);

    return result;
}


bool filesEqual(const FileSummary& summary, const FileTag& tag)
{
    return (summary.filename == tag.filename)
            && (summary.uncompressedMD5 == tag.md5file)
            && (summary.compressedMD5 == tag.md5cfile);
}


bool workflowstepsEqual(const Workflowstep&, const SectionTag&)
{
    // not implemented, yet
    return true;
}


int distance(const FileTag& tag, const FileSummary& summary)
{
    const int FILENAME_DIST = 20;
    const int FILETYPE_DIST = 1;
    const int COMMENT_DIST = 1;
    const int DESCRIPTION_DIST = 1;
    const int CMD5_DIST = 2;
    const int MD5_DIST = 20;
    const int CSIZE_DIST = 2;
    const int SIZE_DIST = 20;
    const int COMPRESS_DIST = 2;
    
    int dist = 0;
    dist += (tag.filename == summary.filename) ? 0 : FILENAME_DIST;
    dist += (tag.filetype == summary.filetype) ? 0 : FILETYPE_DIST;
    dist += (tag.comment  == summary.comment)  ? 0 : COMMENT_DIST;
    dist += (tag.description == summary.description) ? 0 : DESCRIPTION_DIST;
    dist += (tag.md5file == summary.uncompressedMD5) ? 0 : MD5_DIST;
    dist += (tag.filesize == summary.uncompressedSize) ? 0 : SIZE_DIST;
    dist += (tag.md5cfile == summary.compressedMD5) ? 0 : CMD5_DIST;
    dist += (tag.cfilesize == summary.compressedSize) ? 0 : CSIZE_DIST;
    dist += (tag.compression == summary.compressedMethod) ? 0 : COMPRESS_DIST;
    return dist;
}


Summary augmentSummary(const Summary& summary, const SectionTag& section,
        TagHashTable& filetags)
{
    Summary augmentedSummary = summary;
    augmentedSummary.step.tagID = section.id;
    
    // iterate over all files in the summary and search for matching tags
    for (int i = 0; i < augmentedSummary.step.files.count(); i++)
    {
        FileSummary& fsummary = augmentedSummary.step.files[i];
        if (fsummary.embedded)
        {
            // find nearest occurence of file description in tags
            QString hashValue = makeFileHash(fsummary, section.id);
            
            int nearestDist = 1000000;
            TagHashTable::iterator nearestTag = filetags.end();
            
            TagHashTable::iterator it = filetags.find(hashValue);
            while (it != filetags.end() && it.key() == hashValue)
            {
                int dist = distance(it->first, fsummary);
                if (dist < nearestDist)
                {
                    nearestDist = dist;
                    nearestTag = it;
                }
                ++it;
            }
            
            // identify file description with tag and remove the tag from the
            // list of candidates
            if (nearestDist < FILEDIST_THRESHOLD)
            {
                fsummary.tagID = nearestTag->first.id;
                filetags.erase(nearestTag);
            }
        }
    }
    
    // add file summary for every tag which did not match an entry in the
    // workflowstep summary
    //qDebug() << "section id" << section.id;
    //qDebug() << "#unmatched tags:" << filetags.count();
    TagHashTable::iterator it;
    for (it = filetags.begin(); it != filetags.end(); it++)
    {
        if (it->second == summary.step.tagID)
        {
            qDebug() << "foundMismatch";
            FileSummary fsummary = makeFileSummaryFromFileTag(it->first);
            printFileSummary(fsummary);
            augmentedSummary.step.files.append(fsummary);
        }
    }
    
    return augmentedSummary;
}


Summary makeUnmatchedTagSummary(const FileTagList& unmatched)
{
    Summary result;
    result.step.title = "<unmatched files>";
    for (int i = 0; i < unmatched.count(); i++)
    {
        FileTag tag = unmatched[i];
        result.step.files.append(makeFileSummaryFromFileTag(tag));
    }
    return result;
}

