/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: summaryreader.cpp 2677 2013-04-16 10:07:33Z ahuesgen $
 */


#include "../../include/unvhistlib/summaryreader.h"
#include "../../include/unvhistlib/xmlit.h"

#include <QHash>
#include <QtDebug>

// helperfunctions

using namespace vhist;

int str2int(const QString& str)
{
    bool ok = false;
    int res = str.toInt(&ok);
    return ok ? res : -1;
}


bool str2bool(const QString& str)
{
    return bool(str.toInt());
}


/**
 * Creates a new SummaryReader object.
 */
SummaryReader::SummaryReader()
{
}


/**
 * Destroys the SummaryReader object.
 */
SummaryReader::~SummaryReader()
{
}


/**
 * Reads a summary file and creates a Summary object including all
 * the information contained in the file.
 * 
 * @param[in] data The content of the summary file.
 * 
 * @return A Summary object. On success, the \a summaryOK member of the
 * object will be set to \c true.
 */
Summary SummaryReader::readSummary(const QString& data)
{
    VXMLLevel1Iter xi;
    
    for (xi = data; xi; ++xi)
    {
        if (xi.isTag("vhxchange"))
            return parseSummary(xi.getData());
    }
    
    // !sv needs Error checking, should not get here!
    Summary summary;
    return summary;
}


Attribute SummaryReader::parseAttribute(const QString& content)
{
    VXMLLevel1Iter xi;
    Attribute attr;

    for (xi = content; xi; ++xi)
    {
        if (xi.isTag("Key"))
            attr.first = xi.getData();
        else if (xi.isTag("Value"))
            attr.second = xi.getData();
        else if (xi.isTag("Unit"))
        {
            QSharedPointer<QString> unit(new QString(xi.getData()));
            attr.third = unit;
        }
        else
            qDebug() << "unknown tag (parseAttribute):" << xi.getTag() << ":" << xi.getData();
    }
    return attr;
}


Tripple<UserAttributes,UserAttributeOrder,UserAttributeUnits>
        SummaryReader::parseUserAttributes(const QString& content)
{
    VXMLLevel1Iter xi;
    UserAttributes attrs;
    UserAttributeOrder order;
    UserAttributeUnits units;
    
    for (xi = content; xi; ++xi)
    {
        if (xi.isTag("Attribute"))
        {
            Attribute attr = parseAttribute(xi.getData());
            attrs.insert(attr.first, attr.second);
            if (attr.third != NULL)
                units.insert(attr.first, *(attr.third));
            order.append(attr.first);
        }
        else
            qDebug() << "unknown tag (parseUserAttributes):" << xi.getTag() << ":" << xi.getData();
    }
    return Tripple<UserAttributes, UserAttributeOrder, UserAttributeUnits>
            (attrs, order, units);

}


Workflowstep SummaryReader::parseWorkflowstep(const QString& content)
{
    VXMLLevel1Iter xi;
    Workflowstep step;
    for (xi = content; xi; ++xi)
    {
        if (xi.isTag("VhistRootFile"))
            step.vhistRootFile = xi.getData();
        else if (xi.isTag("Timestamp"))
            step.timestamp = xi.getData();
        else if (xi.isTag("Title"))
            step.title = xi.getData();
        else if (xi.isTag("Description"))
            step.description = xi.getData();
        else if (xi.isTag("Comment"))
            step.comment = xi.getData();
        else if (xi.isTag("Command"))
            step.command = xi.getData();
        else if (xi.isTag("Tool"))
            step.tool = xi.getData();
        else if (xi.isTag("ToolPath"))
            step.toolPath = xi.getData();
        else if (xi.isTag("Host"))
            step.host = xi.getData();
        else if (xi.isTag("User"))
            step.user = xi.getData();
        else if (xi.isTag("Creator"))
            step.creator = xi.getData();
        else if (xi.isTag("PDFTitle"))
            step.pdfTitle = xi.getData();
        else if (xi.isTag("PDFAuthor"))
            step.pdfAuthor = xi.getData();
        else if (xi.isTag("PDFSubject"))
            step.pdfSubject = xi.getData();
        else if (xi.isTag("PDFKeywords"))
            step.pdfKeywords = xi.getData();
        else if (xi.isTag("PDFProducer"))
            step.pdfProducer = xi.getData();
        else if (xi.isTag("PDFCreator"))
            step.pdfCreator = xi.getData();
        else if (xi.isTag("InFile"))
        {
            FileSummary f = parseFile(xi.getData());
            f.filePurpose = "InFile";
            step.files.append(f);
        }
        else if (xi.isTag("OutFile"))
        {
            FileSummary f = parseFile(xi.getData());
            f.filePurpose = "OutFile";
            step.files.append(f);
        }
        else if (xi.isTag("UserAttributes"))
        {
            Tripple<UserAttributes, UserAttributeOrder, UserAttributeUnits>
                    attrs;
            attrs = parseUserAttributes(xi.getData());
            step.userAttrs = attrs.first;
            step.userAttrOrder = attrs.second;
            step.userAttrUnits = attrs.third;
        }
        else
            qDebug() << "unknown tag (parseWorkflowstep):"
                     << xi.getTag() << ":" << xi.getData();
    }
    step.tagID = -1;
    return step;
}


FileSummary SummaryReader::parseFile(const QString& content)
{
    VXMLLevel1Iter xi;
    FileSummary fileDesc;
    for (xi = content; xi; ++xi)
    {
        if (xi.isTag("Filename"))
            fileDesc.filename = xi.getData();
        else if (xi.isTag("Filepath"))
            fileDesc.filepath = xi.getData();
        else if (xi.isTag("Filetype"))
            fileDesc.filetype = xi.getData();
        else if (xi.isTag("Description"))
            fileDesc.description = xi.getData();
        else if (xi.isTag("Comment"))
            fileDesc.comment = xi.getData();
        else if (xi.isTag("LastModified"))
            fileDesc.lastModified = xi.getData();
        else if (xi.isTag("GenerateMD5"))
            fileDesc.generateMD5 = str2bool(xi.getData());
        else if (xi.isTag("Embedded"))
            fileDesc.embedded = str2bool(xi.getData());
        else if (xi.isTag("UncompressedSize"))
            fileDesc.uncompressedSize = str2int(xi.getData());
        else if (xi.isTag("UncompressedMD5"))
            fileDesc.uncompressedMD5 = xi.getData();
        else if (xi.isTag("CompressedMethod"))
            fileDesc.compressedMethod = xi.getData();
        else if (xi.isTag("CompressedSize"))
            fileDesc.compressedSize = str2int(xi.getData());
        else if (xi.isTag("CompressedMD5"))
            fileDesc.compressedMD5 = xi.getData();
        else if (xi.isTag("Preview"))
            fileDesc.preview = str2bool(xi.getData());
        else if (xi.isTag("PreviewWS"))
            fileDesc.previewWS = str2bool(xi.getData());
        else if (xi.isTag("Thumbnail"))
            fileDesc.thumbnail = str2bool(xi.getData());
        else if (xi.isTag("ThumbnailOnly"))
            fileDesc.thumbnailOnly = str2bool(xi.getData());
        else if (xi.isTag("UserAttributes"))
        {
            Tripple<UserAttributes, UserAttributeOrder, UserAttributeUnits> attrs;
            attrs = parseUserAttributes(xi.getData());
            fileDesc.userAttrs = attrs.first;
            fileDesc.userAttrOrder = attrs.second;
            fileDesc.userAttrUnits = attrs.third;
        }
        else
            qDebug() << "unknown tag (ParseFile):" << xi.getTag() << ":" << xi.getData();
    }
    fileDesc.tagID = -1;
    fileDesc.unmatchedTag = false;
    return fileDesc;
}


Summary SummaryReader::parseSummary(const QString& content)
{
    VXMLLevel1Iter xi;
    Summary summary;
    for (xi = content; xi; ++xi)
    {
        if (xi.isTag("Workflowstep"))
            summary.step = parseWorkflowstep(xi.getData());
        else if (xi.isTag("MD5Signature"))
            summary.md5 = xi.getData();
    }
    summary.summaryOK = true;
    return summary;
}


void vhist::printFileSummary(const vhist::FileSummary& f)
{
    qDebug() << "  TagID" << f.tagID;
    qDebug() << "  Filename" << f.filename;
    qDebug() << "  Filepath" << f.filepath;
    qDebug() << "  Filetype" << f.filetype;
    qDebug() << "  Description" << f.description;
    qDebug() << "  Comment" << f.comment;
    qDebug() << "  LastModified" << f.lastModified;
    qDebug() << "  GenerateMD5" << f.generateMD5;
    qDebug() << "  Embedded" << f.embedded;
    qDebug() << "  UncompressedSize" << f.uncompressedSize;
    qDebug() << "  UncompressedMD5" << f.uncompressedMD5;
    qDebug() << "  CompressedMethod" << f.compressedMethod;
    qDebug() << "  CompressedSize" << f.compressedSize;
    qDebug() << "  CompressedMD5" << f.compressedMD5;
    qDebug() << "  FilePurpose" << f.filePurpose;
    qDebug() << "  Preview" << f.preview;
    qDebug() << "  PreviewWS" << f.previewWS;
    qDebug() << "  Thumbnail" << f.thumbnail;
    qDebug() << "  ThumbnailOnly" << f.thumbnailOnly;
    qDebug() << "  File User Attributes:";
    qDebug() << "   " << f.userAttrs;
}


void printWorkflowstep(const Workflowstep& s)
{
    qDebug() << "TagID" << s.tagID;
    qDebug() << "VhistRootFile:" << s.vhistRootFile;
    qDebug() << "Timestamp:" << s.timestamp;
    qDebug() << "Title:" << s.title;
    qDebug() << "Description:" << s.description;
    qDebug() << "Comment:" << s.comment;
    qDebug() << "Command:" << s.command;
    qDebug() << "Tool:" << s.tool;
    qDebug() << "ToolPath:" << s.toolPath;
    qDebug() << "Host:" << s.host;
    qDebug() << "User:" << s.user;
    qDebug() << "Creator:" << s.creator;
    qDebug() << "PDFTitle:" << s.pdfTitle;
    qDebug() << "PDFAuthor:" << s.pdfAuthor;
    qDebug() << "PDFSubject:" << s.pdfSubject;
    qDebug() << "PDFKeywords:" << s.pdfKeywords;
    qDebug() << "PDFProducer:" << s.pdfProducer;
    qDebug() << "PDFCreator:" << s.pdfCreator;
    
    
    for (int i = 0; i < s.files.count(); i++)
    {
        printFileSummary(s.files.at(i));
    }
    
    qDebug() << "Step User Attributes:";
    qDebug() << " " << s.userAttrs;
}

void vhist::printSummary(const Summary& s)
{
    printWorkflowstep(s.step);
    qDebug() << "Summary MD5:" << s.md5;
}


QString vhist::makeFileHash(const FileSummary& f, int sectionID)
{
    QString hashString("file: <%1>-<%2>-<%3>-<%4>");
    QString idString = QString::number(sectionID);
    QString sizeString = QString::number(f.uncompressedSize);
    
    return hashString.arg(idString, f.filename, f.uncompressedMD5, sizeString);
}


Summary vhist::makeSummaryFromTags(const SectionTag& tag, const FileTagList& files)
{
    Summary result;
    result.step.tagID = tag.id;
    result.step.title = tag.title;
    result.step.creator = tag.creator;
    
    for (int i = 0; i < files.count(); i++)
    {
        FileTag fileTag = files.at(i);
        FileSummary fsummary;
        fsummary.tagID = fileTag.id;
        fsummary.filetype = fileTag.filetype;
        fsummary.filename = fileTag.filename;
        fsummary.description = fileTag.description;
        fsummary.comment = fileTag.comment;
        fsummary.compressedMethod = fileTag.compression;
        fsummary.compressedSize = fileTag.cfilesize;
        fsummary.uncompressedSize = fileTag.filesize;
        fsummary.compressedMD5 = fileTag.md5cfile;
        fsummary.uncompressedMD5 = fileTag.md5file;
        
        result.step.files.append(fsummary);
    }
    return result;
}


FileSummary vhist::makeFileSummaryFromFileTag(const FileTag& tag)
{
    FileSummary result;
    result.tagID = tag.id;
    result.filetype = tag.filetype;
    result.filename = tag.filename;
    result.description = tag.description;
    result.comment = tag.comment;
    result.compressedMethod = tag.compression;
    result.compressedSize = tag.cfilesize;
    result.uncompressedSize = tag.filesize;
    result.compressedMD5 = tag.md5cfile;
    result.uncompressedMD5 = tag.md5file;
    return result;
}

