/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: validationqueue.cpp 2179 2011-06-24 09:03:24Z ahuesgen $
 */

#include "../../include/unvhistlib/validationqueue.h"
#include "../../include/unvhistlib/vhistparser.h"

#include <QtDebug>

using namespace vhist;

/**
 * Creates a ValidationQueue object.
 *
 * @param[in] parser The parser, who backs the validation queue.
 * All files and sections are taken from this VhistParser object.
 * Unless you only want to validate extracted files on your harddisk,
 * \a parser should not be \c NULL.
 *
 * @param[in] chunksize The maximum amount of data in bytes which is processed
 * during one call of ValidationQueue::process().
 */
ValidationQueue::ValidationQueue(VhistParser* parser, int chunksize) :
        m_totalBytes(0), m_chunksize(chunksize)
{
    m_chunksize = m_chunksize < 1 ? 100000 : m_chunksize;
    m_stopwatch.publishToLogSignal(this);
    reinit(parser);
}


/**
 * Destroys the ValidationQueue object.
 */
ValidationQueue::~ValidationQueue()
{
}


void ValidationQueue::reinit(VhistParser* parser)
{
    m_validatees.clear();
    m_parser = parser;
    m_totalBytes = 0;
    m_totalHandled = 0;
}


/**
 * Clears and reinitializes the ValidationQueue.
 * 
 * @param[in] parser The parser which is used as datasource for this
 * ValidationQueue.
 */
void ValidationQueue::reset(VhistParser* parser)
{
    reinit(parser);
    m_stopwatch.start("validation finished");
}


/**
 * Resets the ValidationQueue and enqueues all files and sections from
 * the given \a parser.
 *
 * @param[in] parser The parser which is used as datasource for this
 * ValidationQueue.
 */
void ValidationQueue::repopulate(VhistParser* parser)
{
    reset(parser);
    if (m_parser != NULL)
    {
        FileTagList files = m_parser->getFiles();
        SectionTagList sections = m_parser->getSections();
        
        for (int i = sections.count() - 1; i >= 0; i--)
        {
            Validatee v = validateeFromSection(sections.at(i));
            m_validatees.enqueue(v);
            m_totalBytes += v.numBytes;
        }
        
        for (int i = 0; i < files.count(); i++)
        {
            //qDebug() << "file: " << i;
            Validatee v = validateeFromFile(files.at(i));
            m_validatees.enqueue(v);
            m_totalBytes += v.numBytes;
        }
    }
    m_bytesHandled = 0;
}


/**
 * Stops the validation and removes all validatees from the queue.
 */
void ValidationQueue::cancelValidation()
{
    m_validatees.clear();
    m_bytesHandled = 0;
    m_totalBytes = 0;
    m_totalHandled = 0;
    m_stopwatch.stop();
}


/**
 * Adds a list of files and sections to the validation queue. The files and
 * sections are taken from the VhistParser object associated with the
 * queue.
 *
 * @param[in] fileIDs A list of ids of embedded files in the associated
 * VHIST file, which should be appended to the queue.
 *
 * @param[in] sectionIDs A list of ids of sections in the associated VHIST
 * file, which should be appended to the queue.
 */
void ValidationQueue::addValidatees(QList<int> fileIDs, QList<int> sectionIDs)
{
    if (m_parser != NULL)
    {
        FileTagList files = m_parser->getFiles();
        SectionTagList sections = m_parser->getSections();
        
        for (int i = sections.count() - 1; i >= 0; i--)
        {
            const SectionTag& s = sections.at(i);
            if (s.id != -1 && sectionIDs.contains(s.id))
            {
                Validatee v = validateeFromSection(s);
                m_validatees.enqueue(v);
                m_totalBytes += v.numBytes;
            }
        }
        
        for (int i = 0; i < files.count(); i++)
        {
            const FileTag& f = files.at(i);
            if (f.id != -1 && fileIDs.contains(f.id))
            {
                Validatee v = validateeFromFile(f);
                m_validatees.enqueue(v);
                m_totalBytes += v.numBytes;
            }
        }
    }
}


/**
 * Adds a list of files in the filesystem to the validation queue.The files
 * are characterized by their filenames, their filesize and the expected
 * md5 sum.
 *
 * @param[in] files The list of files which should be added to the queue.
 */
void ValidationQueue::addValidatees(QList<SavedFile> files)
{
    for (int i = 0; i < files.count(); i++)
    {
        Validatee v = validateeFromSavedFile(files.at(i));
        m_validatees.enqueue(v);
        m_totalBytes += v.numBytes;
    }
}


/**
 * Continues processing entries in the queue. The queue will only process
 * a limited amount of data during one call. The size of each chunk can be
 * specified in the constructor. Small sizes result in high responsiveness of
 * 
 */
bool ValidationQueue::process()
{
    QString currentName;
    if (!m_validatees.empty())
    {
        Validatee v = m_validatees.head();
        currentName = v.name;
        
        if (m_bytesHandled == 0)
        {
            m_validator.reinit(v.md5, v.cmd5);
            if (v.type == VALIDATEE_SAVEDFILE)
            {
                m_currentFile.close();
                m_currentFile.setFileName(v.name);
                m_currentFile.open(QIODevice::ReadOnly);
            }
        }
        
        int numBytes = qMin(v.numBytes - m_bytesHandled, m_chunksize);
        if (v.type == VALIDATEE_SECTION && v.start + m_bytesHandled < v.tagPos)
            numBytes = qMin(numBytes, v.tagPos - (v.start + m_bytesHandled));
        
        QByteArray bytes;
        if (v.type == VALIDATEE_SAVEDFILE)
            bytes = m_currentFile.read(numBytes);
        else
            bytes = m_parser->getBytes(v.start + m_bytesHandled, numBytes);
        
        m_validator.update(bytes);
        m_bytesHandled += numBytes;
        m_totalHandled += numBytes;
        
        // if validatee is a section, we need to replace the checksum in the tag
        if (v.type == VALIDATEE_SECTION && v.tagPos == v.start + m_bytesHandled)
        {
            m_validator.update(v.replacementTag);
            m_bytesHandled += v.replacementTag.size();
        }
        
        if (m_bytesHandled == v.numBytes)
        {
            if (v.type == VALIDATEE_FILE)
            {
                //qDebug() << "file: " << v.id;
                emit fileProcessed(v.id, m_validator.isValid());
            }
            else if (v.type == VALIDATEE_SECTION)
            {
                //qDebug() << "section: " << v.id;
                emit sectionProcessed(v.id, m_validator.isValid());
            }
            else if (v.type == VALIDATEE_SAVEDFILE)
            {
                //qDebug() << "file on harddisk:" << v.name;
                emit savedFileProcessed(v.name, m_validator.isValid());
                if (m_validator.isValid())
                    emit log(tr("Checksum of file %1 is correct").arg(v.name));
                else
                    emit log(tr("Checksum of file %1 is incorrect").arg(v.name));
            }
            
            nextEntry();
        }
    }
    
    // calculate percentage finished
    int percent = static_cast<int>(100.0 * m_totalHandled / m_totalBytes);
    emit percentFinishedChanged(percent, currentName);
    
    if (m_validatees.empty())
    {
        m_stopwatch.stop();
        m_bytesHandled = 0;
        m_totalBytes = 0;
        m_totalHandled = 0;
    }
    
    return !m_validatees.empty();
}


void ValidationQueue::nextEntry()
{
    if (!m_validatees.empty())
        m_validatees.dequeue();
    m_bytesHandled = 0;
    m_currentFile.close();
}


Validatee ValidationQueue::validateeFromFile(const FileTag& tag)
{
    Validatee v;
    v.type = VALIDATEE_FILE;
    v.name = tag.filename;
    v.id = tag.id;
    v.start = tag.position + tag.taglen + tag.offset;
    v.md5 = tag.md5file;
    v.tagPos = -1;
    
    if (tag.compression == "flate")
    {
        v.numBytes = tag.cfilesize;
        v.cmd5 = tag.md5cfile;
    }
    else
    {
        v.numBytes = tag.filesize;
        v.cmd5 = "";
    }
    return v;
}


Validatee ValidationQueue::validateeFromSection(const SectionTag& tag)
{
    Validatee v;
    v.type = VALIDATEE_SECTION;
    v.name = tag.title == "" ? "<no title>" : tag.title;
    v.id = tag.id;
    v.start = tag.position - tag.left;
    v.tagPos = tag.position;
    v.replacementTag = tag.validationTag;
    v.numBytes = tag.size;
    v.md5 = tag.md5section;
    v.cmd5 = "";
    return v;
}


Validatee ValidationQueue::validateeFromSavedFile(const SavedFile& file)
{
    Validatee v;
    v.type = VALIDATEE_SAVEDFILE;
    v.md5 = file.md5;
    v.cmd5 = "";
    v.name = file.filename;
    v.numBytes = file.filesize;
    return v;
}


// documentation of Qt signals

/**
 * @fn void ValidationQueue::fileProcessed(int id, bool valid)
 * 
 * This signal is emitted each time a file has been completely processed.
 * 
 * @param[in] id The id of the processed file.
 * @param[in] valid \c true if the file's checksum is correct.
 */

/**
 * @fn void ValidationQueue::sectionProcessed(int id, bool valid)
 * 
 * This signal is emitted each time a file has been completely processed.
 * 
 * @param[in] id The id of the processed file.
 * @param[in] valid \c true if the section's checksum is correct.
 */

/**
 * @fn void ValidationQueue::savedFileProcessed(const QString filename, bool valid)
 * 
 * This signal is emitted each time a file on the harddisk has been
 * completely processed.
 * 
 * @param[in] filename The name of the processed file.
 * @param[in] valid \c true if the files's checksum is correct.
 */


/**
 * @fn void ValidationQueue::percentFinishedChanged(int percent, const QString filename)
 * 
 * This signal is emitted each time the percentage of validated data in the
 * queue changes.
 * 
 * @param[in] percent The percentage of data from the queue, which has
 * already been validated.
 * 
 * @param[in] filename The name of the file which is processed at the
 * moment.
 */

