/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: validator.cpp 2179 2011-06-24 09:03:24Z ahuesgen $
 */

#include "../../include/unvhistlib/validator.h"

#include <QByteArray>
#include <QtDebug>


using namespace vhist;

/**
 * Creates a new Validator object. The MD5 sums passed as arguments are
 * the checksums you expect the stream to have.
 * 
 * @param[in] md5 The expected md5 sum of the uncompressed stream.
 * @param[in] compressedMD5 The expected md5 sum of the compressed stream.
 * If the stream is not compressed, pass an empty string as \a compressedMD5.
 * This will deactivate the inflation.
 */
Validator::Validator(const QString& md5, const QString& compressedMD5)
{
    reinit(md5, compressedMD5);
}


/**
 * Destroys the Validator object.
 */
Validator::~Validator()
{
}


/**
 * Resets the Validator. This function can be used to reuse the same
 * Validator object for several datastreams.
 * 
 * @param[in] md5 The expected md5 sum of the uncompressed stream.
 * @param[in] compressedMD5 The expected md5 sum of the compressed stream.
 * If the stream is not compressed, pass an empty string as \a compressedMD5.
 * This will deactivate the inflation.
 */
void Validator::reinit(const QString& md5, const QString& compressedMD5)
{
    m_expectedMD5 = md5;
    m_expectedCompressedMD5 = compressedMD5;
    m_decompress = (m_expectedCompressedMD5 != "");
    
    // init check sum generators and inflater
    m_md5.reset();
    m_compressedMD5.reset();
    m_inflater.reset();
}


/**
 * Validates the next consecutive part of the datastream. The data
 * will be decompressed if neccessary and the md5 sums are updated.
 * 
 * You can call isValid() after feeding the Validator with the
 * complete datastream to check if the md5 sums matched and the
 * compressed stream has been consistent.
 * 
 * @param[in] data The next part of the datastream.
 */
void Validator::update(const QByteArray& data)
{
    if (m_decompress)
    {
        m_compressedMD5.update(data);
        QByteArray decompressedData = m_inflater.inflateData(data);
        m_md5.update(decompressedData);
    }
    else
        m_md5.update(data);
}


/**
 * Returns \c true if the validation of the datastream has been successful.
 * The reason for a failed validation can be a corrupted compressed
 * stream or a mismatch of MD5 sums.
 *
 * @return Whether the validation has been successful.
 */
bool Validator::isValid()
{
    if (m_decompress && (!m_inflater.hasFinished() || !m_inflater.isValid()))
        return false;
    
    if (m_expectedMD5 != m_md5.getHexDigest())
        return false;
    
    if (m_decompress &&
            m_expectedCompressedMD5 != m_compressedMD5.getHexDigest())
        return false;
    
    return true;
}

