/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: pathexpander.cpp 2179 2011-06-24 09:03:24Z ahuesgen $
 */

#include "../../include/vhistdifflib/pathexpander.h"
#include "../../include/vhistdifflib/costmatrix.h"

#include <QHash>

using namespace vhist;

PathExpander::PathExpander(const WorkflowTree* tree1, const WorkflowTree* tree2,
        const CostMatrix* costs)
        : m_tree1(tree1), m_tree2(tree2), m_costs(costs)
{
}


PathExpander::~PathExpander()
{
}


float PathExpander::costForAssociation(NodeId node1, NodeId node2,
        const Associations& associations,
        const Associations& revAssociations) const
{
    float costs = m_costs->costs(node1, node2);
    NodeId parent1 = m_tree1->parent(node1);
    NodeId parent2 = m_tree2->parent(node2);
    // TODO: check again
    if (node1 != NULL_NODE && node2 != NULL_NODE &&
            associations.value(parent1, NULL_NODE) != NULL_NODE &&
            revAssociations.value(parent2, NULL_NODE) != NULL_NODE &&
            (parent1 == INVALID_NODE || parent2 == INVALID_NODE ||
            associations[parent1] != parent2))
        costs += m_costs->costsWrongParent();
    return costs;
}


bool PathExpander::calcParentCostsLater(NodeId node1, NodeId node2,
        const Associations& associations,
        const Associations& revAssociations) const
{
    NodeId parent1 = m_tree1->parent(node1);
    NodeId parent2 = m_tree2->parent(node2);
    
    return (associations[parent1] == NULL_NODE ||
            revAssociations[parent2] == NULL_NODE);
}


float PathExpander::costForUnmatchedNodes(int numOfUnmatchedNodes) const
{
    return m_costs->costsUnmatchedNode() * numOfUnmatchedNodes;
}


QLinkedList<NodeId> PathExpander::nodeIds(TreeId treeId) const
{
    if (treeId == FIRST_TREE)
        return m_tree1->nodeIdsLinked();
    else if (treeId == SECOND_TREE)
        return m_tree2->nodeIdsLinked();
    else
        return QLinkedList<NodeId>();
}


bool PathExpander::parentChanged(NodeId node1,
        const Associations& associations,
        const Associations& revAssociations) const
{
    NodeId node2 = associations[node1];
    if (node1 == INVALID_NODE || node2 == INVALID_NODE)
        return false;

    NodeId ancestor1 = m_tree1->parent(node1);
    while (associations[ancestor1] == NULL_NODE)
        ancestor1 = m_tree1->parent(ancestor1);

    NodeId ancestor2 = m_tree2->parent(node2);
    while (revAssociations[ancestor2] == NULL_NODE)
        ancestor2 = m_tree2->parent(ancestor2);

    return associations[ancestor1] != ancestor2;
}


float PathExpander::calcMissingParentCosts(const Associations& associations,
        const Associations& revAssociations,
        const QHash<NodeId, bool>& parentCostsMissing) const
{
    float costs = 0;
    QList<NodeId> keys = associations.keys();
    for (int i = 0; i < keys.size(); ++i)
    {
        NodeId node1 = keys[i];
        if (parentCostsMissing[i])
            if (parentChanged(node1, associations, revAssociations))
                costs += m_costs->costsWrongParent();
    }
    return costs;
}


/**
    Calculates heuristic costs for the remaining associations (used in A*).
    
    Searches for the best association for each node in list \c short and
    accumulates the costs. Here, \c short is the shorter one of both lists.
    COST_UNMATCHED_NODE * (long - short) is added to reflect the costs created
    by unassociated nodes. Wrong parents or dublicate associations are not
    penalized, resulting in an underestimation of the actual costs.
**/
float PathExpander::heuristicCosts(const QLinkedList<NodeId>& nodes1,
        const QLinkedList<NodeId>& nodes2) const
{
    QLinkedList<NodeId> shortList;
    QLinkedList<NodeId> longList;
    bool firstShorter = nodes1.size() < nodes2.size();
    shortList = firstShorter? nodes1 : nodes2;
    longList = firstShorter? nodes2 : nodes1;
    
    float cumcosts = costForUnmatchedNodes(longList.size() - shortList.size());
    for (QLinkedList<NodeId>::Iterator it = shortList.begin();
        it != shortList.end(); ++it)
    {
        float mincosts = m_costs->costsUnmatchedNode();
        for (QLinkedList<NodeId>::Iterator it2 = longList.begin();
            it2 != longList.end(); ++it2)
        {
            float newcosts = 0;
            if (firstShorter)
                newcosts = m_costs->costs(*it, *it2);
            else
                newcosts = m_costs->costs(*it2, *it);
            mincosts = qMin(mincosts, newcosts);
        }
        cumcosts += mincosts;
    }
    return cumcosts;
}
