/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: workflowtree.cpp 2179 2011-06-24 09:03:24Z ahuesgen $
 */

#include "../../include/vhistdifflib/workflowtree.h"

#include "../../include/unvhistlib/summarycreator.h"
#include "../../include/unvhistlib/extractionqueue.h"

#include <QDir>
#include <QFile>
#include <QLinkedList>
#include <QTemporaryFile>
#include <QtDebug>

using namespace vhist;

WorkflowTree::WorkflowTree(QObject* listener)
    : LogSource(listener)
{
    m_filemanager.publishToLogSignal(this);
    clear();
}


WorkflowTree::WorkflowTree(const WorkflowTree& rhs) : LogSource()
{
    *this = rhs;
}


WorkflowTree::~WorkflowTree()
{
}


WorkflowTree& WorkflowTree::operator=(const WorkflowTree& rhs)
{
    if (this != &rhs)
    {
        m_nextId = rhs.m_nextId;
        m_nodes = rhs.m_nodes;
        m_children = rhs.m_children;
        m_filemanager = rhs.m_filemanager;
    }

    return *this;
}


NodeId WorkflowTree::addTreeNode(NodeId parentId, const Workflowstep& wfs,
        const QString& vhistfile, const QString& tmpfile)
{
    addTreeNode(WorkflowTreeNode(m_nextId, parentId, wfs, vhistfile,
            tmpfile));
    return m_nextId++;
}


void WorkflowTree::addTreeNode(const WorkflowTreeNode& node)
{
    m_nodes.insert(node.nodeId(), node);

    if (!m_children.contains(node.parentId()))
        m_children.insert(node.parentId(), QList<NodeId>());

    m_children[node.parentId()].append(node.nodeId());
}


void WorkflowTree::clear()
{
    m_nextId = 0;
    m_nodes.clear();
    m_children.clear();
    m_filemanager.clear();
}


const Workflowstep* WorkflowTree::workflowStep(NodeId nodeId) const
{
    if (m_nodes.contains(nodeId))
        return const_cast<WorkflowTree*>(this)->m_nodes[nodeId].workflowstep();
    else
        return NULL;
}


NodeId WorkflowTree::parent(NodeId nodeId) const
{
    if (m_nodes.contains(nodeId))
        return m_nodes[nodeId].parentId();
    else
        return INVALID_NODE;
}


QString WorkflowTree::vhistfile(NodeId nodeId) const
{
    if (m_nodes.contains(nodeId))
        return m_nodes[nodeId].vhistfile();
    else
        return QString();
}


QString WorkflowTree::fileOnDisk(NodeId nodeId) const
{
    if (m_nodes.contains(nodeId))
        return m_nodes[nodeId].fileOnDisk();
    else
        return QString();
}


QList<NodeId> WorkflowTree::children(NodeId nodeId) const
{
    if (m_nodes.contains(nodeId))
        return m_children[nodeId];
    else
        return QList<NodeId>();
}


SPVhistParser WorkflowTree::vhistParser(NodeId nodeId)
{
    QString filename = fileOnDisk(nodeId);
    return m_filemanager.parser(filename);
}


QList<NodeId> WorkflowTree::nodeIds() const
{
    QList<NodeId> result;
    for (TreeNodeHash::const_iterator it = m_nodes.constBegin();
            it != m_nodes.constEnd(); ++it)
        result << it.key();
    return result;
}


QLinkedList<NodeId> WorkflowTree::nodeIdsLinked() const
{
    QLinkedList<NodeId> result;
    for (TreeNodeHash::const_iterator it = m_nodes.constBegin();
            it != m_nodes.constEnd(); ++it)
        result << it.key();
    return result;
}


QList<NodeId> WorkflowTree::serializedTree() const
{
    NodeId rootNode = root();
    if (rootNode != INVALID_NODE)
        return serializeFromRoot(rootNode);
    else
        return QList<NodeId>();
}


NodeId WorkflowTree::root() const
{
    for (TreeNodeHash::const_iterator it = m_nodes.constBegin();
    it != m_nodes.constEnd(); ++it)
        if (it.value().parentId() == NODE_ABOVE_ROOT)
            return it.key();

    return INVALID_NODE;
}


QList<NodeId> WorkflowTree::serializeFromRoot(const NodeId& root) const
{
    QList<NodeId> result;
    QLinkedList<NodeId> nodesToExpand;
    nodesToExpand << root;
    while(!nodesToExpand.empty())
    {
        NodeId currentNode = nodesToExpand.first();
        nodesToExpand.removeFirst();
        result.append(currentNode);
        QList<NodeId> childList = children(currentNode);
        for (int i = 0; i < childList.size(); ++i)
            nodesToExpand << childList.at(i);
    }
    return result;
}


QString WorkflowTree::extractFile(SPVhistParser vhistFile, int tagId)
{
    QString filename;
    {
        QTemporaryFile tempFile;
        if (tempFile.open())
            filename =tempFile.fileName();
        tempFile.setAutoRemove(false);
    }

    if (filename != "")
    {
        // create and fill ExtractionQueue
        vhist::ExtractionQueue queue(vhistFile.data());
        queue.publishToLogSignal(this);
        QHash<int, QString> files;
        files.insert(tagId, filename);
        queue.addExtractees(files);

        // extract file
        bool running = true;
        while (running)
            running = queue.process();
    }
    return filename;
}


void WorkflowTree::fillTree(SPVhistParser vhistFile,
        NodeId parent, const QString& vhistFileName, const QString& fileOnDisk)
{
    vhist::SummaryCreator creator;
    vhist::Summaries summaries = creator.createSummaries(*vhistFile);

    for (int i = summaries.size() - 1; i >= 0; --i)
    {
        vhist::Summary current(summaries.at(i));
        parent = addTreeNode(parent, current.step, vhistFileName,
                fileOnDisk);

        QList<vhist::FileSummary> files = current.step.files;
        for (int j = 0; j < files.size(); ++j)
        {
            vhist::FileSummary cfile = files.at(j);
            if (cfile.filename.endsWith(".vhist") && cfile.embedded)
            {
                QString tmpfilename = extractFile(vhistFile, cfile.tagID);
                if (tmpfilename != "")
                {
                    SPVhistParser parser =
                            m_filemanager.addVhistFile(tmpfilename, true);
                    fillTree(parser, parent, cfile.filename,
                            tmpfilename);
                }
            }
        }
    }
}


void WorkflowTree::treeFromVhistFile(SPVhistParser vhistFile,
        const QString& filename)
{
    clear();
    m_filemanager.addVhistFile(vhistFile, filename);
    fillTree(vhistFile, NODE_ABOVE_ROOT,
            filename, filename);
}


void WorkflowTree::treeFromVhistFile(const QString& vhistFile)
{
    SPVhistParser parser(new vhist::VhistParser(vhist::MODE_ROBUST));
    parser->open(vhistFile);
    treeFromVhistFile(parser, vhistFile);
}
