/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: checkoperatortypes.cpp 2169 2011-05-30 09:11:44Z ahuesgen $
 */

#include "../../include/vhistfindlib/checkoperatortypes.h"

using namespace vhist;

enum FittingTypes
{
    TYPE_FITS_YES = 1,
    TYPE_FITS_MAYBE = 2,
    TYPE_FITS_NO = 4
};


FittingTypes checkIfTypeFits(const QString& expectedType,
        const QString& actualTypes)
{
    if (actualTypes.contains("no" + expectedType))
        return TYPE_FITS_NO;
    else if (actualTypes.contains(expectedType))
        return TYPE_FITS_YES;
    else
        return TYPE_FITS_MAYBE;
}


QString wrongTypeStr(const QString& operand, const QString& type)
{
    return QString("\n'%1' is not a %2.").arg(operand).arg(type);
}


QString maybeWrongTypeStr(const QString& operand, const QString& type)
{
    return QString("\n'%1' may not be a %2.").arg(operand).arg(type);
}


void createErrorOneType(SPExpressionTreeNode node, const QString& type0,
        const QString& type1, const QString& expectedType, QueryErrors& errors)
{
    ExpressionValue v = node->m_value;
    QString arg0 = node->m_children[0]->m_originalText;
    QString arg1 = node->m_children[1]->m_originalText;
    QString op = v.operatorName();

    FittingTypes fitting0 = checkIfTypeFits(expectedType, type0);
    FittingTypes fitting1 = checkIfTypeFits(expectedType, type1);

    if (fitting0 == TYPE_FITS_NO || fitting1 == TYPE_FITS_NO)
    {
        QString errorMessage("error at pos %1: operator %2 requires "
                "two %3 as arguments.");
        QString e = errorMessage.arg(v.position()).arg(op).arg(expectedType + "s");

        if (fitting0 == TYPE_FITS_NO)
            e += wrongTypeStr(arg0, expectedType);
        if (fitting1 == TYPE_FITS_NO)
            e += wrongTypeStr(arg1, expectedType);

        errors.append(QueryError(e, QueryError::QUERY_ERROR_TYPE_ERROR,
               v.position(), v.originalText()));
    }
    else if (fitting0 == TYPE_FITS_MAYBE || fitting1 == TYPE_FITS_MAYBE)
    {
        QString errorMessage("warning at pos %1: operator %2 requires "
                "two %3 as arguments.");
        QString e = errorMessage.arg(v.position()).arg(op).arg(expectedType + "s");

        if (fitting0 == TYPE_FITS_MAYBE)
            e += maybeWrongTypeStr(arg0, expectedType);
        if (fitting1 == TYPE_FITS_MAYBE)
            e += maybeWrongTypeStr(arg1, expectedType);

        errors.append(QueryError(e, QueryError::QUERY_ERROR_TYPE_ERROR,
               v.position(), v.originalText()));
    }
}


void createErrorTwoTypes(SPExpressionTreeNode node, const QString& type0,
        const QString& type1, const QString& expectedType1,
        const QString& expectedType2, QueryErrors& errors)
{
    ExpressionValue v = node->m_value;
    QString child0 = node->m_children[0]->m_originalText;
    QString child1 = node->m_children[1]->m_originalText;
    QString op = v.operatorName();

    FittingTypes fitting0a = checkIfTypeFits(expectedType1, type0);
    FittingTypes fitting0b = checkIfTypeFits(expectedType2, type0);
    FittingTypes fitting1a = checkIfTypeFits(expectedType1, type1);
    FittingTypes fitting1b = checkIfTypeFits(expectedType2, type1);

    // no solution that works without error
    if ((fitting0a == TYPE_FITS_NO || fitting1a == TYPE_FITS_NO)
            && (fitting0b == TYPE_FITS_NO || fitting1b == TYPE_FITS_NO))
    {
        QString errorMessage("error at pos %1: operator %2 requires "
                "either two %3 or two %4 as arguments.");
        QString e = errorMessage.arg(v.position()).arg(op)
                .arg(expectedType1 + "s").arg(expectedType2 + "s");

        QString xOrY = QString("%1 or %2").arg(expectedType1).arg(expectedType2);
        bool problemFound = false;
        if (fitting0a == TYPE_FITS_NO && fitting0b == TYPE_FITS_NO)
        {
            e += wrongTypeStr(child0, xOrY);
            problemFound = true;
        }
        if (fitting1a == TYPE_FITS_NO && fitting1b == TYPE_FITS_NO)
        {
            e += wrongTypeStr(child1, xOrY);
            problemFound = true;
        }
        if (!problemFound)
        {
            if (fitting0a == TYPE_FITS_NO)
                e += wrongTypeStr(child0, expectedType1);
            else if (fitting0b == TYPE_FITS_NO)
                e += wrongTypeStr(child0, expectedType2);

            if (fitting1a == TYPE_FITS_NO)
                e += wrongTypeStr(child1, expectedType1);
            else if (fitting1b == TYPE_FITS_NO)
                e += wrongTypeStr(child1, expectedType2);
        }
        errors.append(QueryError(e, QueryError::QUERY_ERROR_TYPE_ERROR,
               v.position(), v.originalText()));
    }
    // no solution that works without warning
    else if ((fitting0a != TYPE_FITS_YES || fitting1a != TYPE_FITS_YES)
            && (fitting0b != TYPE_FITS_YES || fitting1b != TYPE_FITS_YES))
    {
        QString errorMessage("warning at pos %1: operator %2 requires "
                "either two numbers or two dates as arguments.");
        QString e = errorMessage.arg(v.position()).arg(op);

        // solution 0 is applicable
        if (fitting0a != TYPE_FITS_NO && fitting1a != TYPE_FITS_NO)
        {
            if (fitting0a == TYPE_FITS_MAYBE)
                e += maybeWrongTypeStr(child0, expectedType1);
            if (fitting1a == TYPE_FITS_MAYBE)
                e += maybeWrongTypeStr(child1, expectedType1);
        }
        // solution 1 is applicable
        if (fitting0b != TYPE_FITS_NO && fitting1b != TYPE_FITS_NO)
        {
            if (fitting0b == TYPE_FITS_MAYBE)
                e += maybeWrongTypeStr(child0, expectedType2);
            if (fitting1b == TYPE_FITS_MAYBE)
                e += maybeWrongTypeStr(child1, expectedType2);
        }

        errors.append(QueryError(e, QueryError::QUERY_ERROR_TYPE_ERROR,
                v.position(), v.originalText()));
    }
}



void vhist::checkLogicalOperators(SPExpressionTreeNode node, const QString& type0,
        const QString& type1, QueryErrors& errors)
{
    createErrorOneType(node, type0, type1, "boolean", errors);
}


void vhist::checkComparissonOperators(SPExpressionTreeNode node, const QString& type0,
        const QString& type1, QueryErrors& errors)
{
    createErrorTwoTypes(node, type0, type1, "number", "date", errors);
}


void vhist::checkStringOperators(SPExpressionTreeNode node, const QString& type0,
        const QString& type1, QueryErrors& errors)
{
    createErrorOneType(node, type0, type1, "string", errors);
}


void vhist::checkNumberOperators(SPExpressionTreeNode node, const QString& type0,
        const QString& type1, QueryErrors& errors)
{
    createErrorOneType(node, type0, type1, "number", errors);
}


void vhist::checkUnaryLogicalOperators(SPExpressionTreeNode node,
        const QString& type0, QueryErrors& errors)
{
    ExpressionValue v = node->m_value;
    QString child0 = node->m_children[0]->m_originalText;
    QString op = v.operatorName();

    FittingTypes fitting0 = checkIfTypeFits("boolean", type0);

    if (fitting0 == TYPE_FITS_NO)
    {
        QString errorMessage("error at pos %1: operator %2 requires "
                "a boolean as argument.");
        QString e = errorMessage.arg(v.position()).arg(op);
        e += wrongTypeStr(child0, "boolean");

        errors.append(QueryError(e, QueryError::QUERY_ERROR_TYPE_ERROR,
               v.position(), v.originalText()));
    }
    else if (fitting0 == TYPE_FITS_MAYBE)
    {
        QString errorMessage("warning at pos %1: operator %2 requires "
                "a boolean as argument.");
        QString e = errorMessage.arg(v.position()).arg(op);
        e += maybeWrongTypeStr(child0, "boolean");

        errors.append(QueryError(e, QueryError::QUERY_ERROR_TYPE_ERROR,
               v.position(), v.originalText()));
    }
}
