/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: checkquerytypes.cpp 2169 2011-05-30 09:11:44Z ahuesgen $
 */

#include "../../include/vhistfindlib/checkquerytypes.h"
#include "../../include/vhistfindlib/returnvalue.h"
#include "../../include/vhistfindlib/checkoperatortypes.h"

#include <QStringList>
#include <QHash>

using namespace vhist;


QHash<QString, QString> getAttributeTypes()
{
    /* comma-separated list of the types string, number, date and/or boolean.
       If a type appears in the list, it means that this field can be converted
       to this type. If a type appears as no<type>, e.g. nobool, it means that
       this type can definitely not be converted to this type. If a type does
       not appear in this field, it means that the type may be convertible
       (converting will generate a warning, but not an error)
    */
    QHash<QString, QString> attributes;
    attributes.insert("hash", "string,nonumber,nodate,noboolean");
    attributes.insert("tagid", "string,number,nodate,noboolean");
    attributes.insert("vhistrootfile", "string,noboolean");
    attributes.insert("timestamp", "string,date,nonumber,noboolean");
    attributes.insert("title", "string,noboolean");
    attributes.insert("description", "string,noboolean");
    attributes.insert("comment", "string,noboolean");
    attributes.insert("command", "string,noboolean");
    attributes.insert("tool", "string,noboolean");
    attributes.insert("toolpath", "string,noboolean");
    attributes.insert("host", "string,noboolean");
    attributes.insert("user", "string,noboolean");
    attributes.insert("creator", "string,noboolean");
    attributes.insert("iswfs", "nostring,nonumber,nodate,boolean");
    attributes.insert("isfile", "nostring,nonumber,nodate,boolean");
    attributes.insert("userattrs", "string,noboolean");

    attributes.insert("filename", "string,noboolean");
    attributes.insert("filepath", "string,noboolean");
    attributes.insert("filetype", "string,noboolean");
    attributes.insert("description", "string,noboolean");
    attributes.insert("comment", "string,noboolean");
    attributes.insert("lastmodified", "string,date,nonumber,noboolean");
    attributes.insert("embedded", "nostring,nonumber,nodate,boolean");
    attributes.insert("compressed", "nostring,nonumber,nodate,boolean");
    attributes.insert("filepurpose", "string,nodate,nonumber,noboolean");
    attributes.insert("isinfile", "nostring,nodate,nonumber,boolean");
    attributes.insert("isoutfile", "nostring,nodate,nonumber,boolean");
    attributes.insert("filesize", "string,number,nodate,noboolean");
    attributes.insert("cfilesize", "string,number,nodate,noboolean");
    attributes.insert("md5", "string,nodate,nonumber,noboolean");
    attributes.insert("cmd5", "string,nodate,nonumber,noboolean");
    attributes.insert("parentid", "string,number,nodate,noboolean");

    attributes.insert("any", "string");

    return attributes;
}


QString getTypesOfString(const ReturnValue& val)
{
    QString strtype = "string,boolean";
    if (val.isNumber())
        strtype += ",number";
    else
        strtype += ",nonumber";
    if (val.isDate())
        strtype += ",date";
    else
        strtype += ",nodate";
    return strtype;
}


QString getTypes(SPExpressionTreeNode tree,
        const QHash<QString, QString>& attrs, QueryErrors& errors)
{
    ExpressionValue v = tree->m_value;
    if (v.isString())
    {
        ReturnValue val = ReturnValue::fromString(v.string());
        return getTypesOfString(val);
    }
    else if (v.isVariable())
    {
        QString attribute = v.variableAttribute();
        if (attribute.startsWith("user:"))
            return "string";
        else if (attrs.contains(attribute))
            return attrs.value(v.variableAttribute(), "string");
        else
        {
            QString errorMessage("error at pos %1: attribute %2 does not exist.");
            QString e = errorMessage.arg(v.position()).arg(attribute);

            errors.append(QueryError(e, QueryError::QUERY_ERROR_TYPE_ERROR,
                    v.position(), v.originalText()));
            return "nostring,noboolean,nonumber,nodate";
        }
    }
    else if (v.isOperator() && v.numArgs() == 2)
    {
        // evaluate both children (recursively)
        SPExpressionTreeNode child0 = tree->m_children[0];
        SPExpressionTreeNode child1 = tree->m_children[1];
        QString type0 = getTypes(child0, attrs, errors);
        QString type1 = getTypes(child1, attrs, errors);

        QString op = v.operatorName();

        QStringList opClass1; opClass1 << "&&" << "||" << "implizit &&";
        QStringList opClass2; opClass2 << "<" << "<=" << ">" << ">=";
        QStringList opClass3; opClass3 << "%contains" << "%matches" << "== << !=";
        QStringList opClass4; opClass4 << "%isapprox";

        if (opClass1.contains(op))
            vhist::checkLogicalOperators(tree, type0, type1, errors);
        else if (opClass2.contains(op))
            vhist::checkComparissonOperators(tree, type0, type1, errors);
        else if (opClass3.contains(op))
            vhist::checkStringOperators(tree, type0, type1, errors);
        else if (opClass4.contains(op))
            vhist::checkNumberOperators(tree, type0, type1, errors);
        return "boolean";
    }
    else if (v.isOperator() && v.numArgs() == 1)
    {
        // evaluate child (recursively)
        SPExpressionTreeNode child0 = tree->m_children[0];
        QString type0 = getTypes(child0, attrs, errors);
        vhist::checkUnaryLogicalOperators(tree, type0, errors);

        return "boolean";
    }
    else
        // TODO: write some useful error handling
        return "boolean";
}


bool vhist::checkQueryTypes(SPExpressionTreeNode tree, QueryErrors& errors)
{
    QHash<QString, QString> hash = getAttributeTypes();

    QString result = getTypes(tree, hash, errors);
    if (result.contains("noboolean") || !result.contains("boolean"))
    {
        QString errorMsg("Result of query %1 is not a boolean value");
        errorMsg = errorMsg.arg(tree->m_originalText);
        errors.append(QueryError(errorMsg, QueryError::QUERY_ERROR_TYPE_ERROR,
                tree->m_pos, tree->m_originalText));
        return false;
    }
    return !containsErrors(errors);
}
