/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: cnf.cpp 2169 2011-05-30 09:11:44Z ahuesgen $
 */

#include "../../include/vhistfindlib/cnf.h"

using namespace vhist;

/******************************************************************
 * The index of a literal within a CNF. Contains the index of the *
 * disjunction as well as the index of the literal within the     *
 * disjunction.                                                   *
 ******************************************************************/

CNFLiteralIndex::CNFLiteralIndex(int disjunctionIndex, int literalIndex) :
        m_disjunctionIndex(disjunctionIndex), m_literalIndex(literalIndex)
{
}


int CNFLiteralIndex::disjunctionIndex() const
{
    return m_disjunctionIndex;
}


int CNFLiteralIndex::literalIndex() const
{
    return m_literalIndex;
}


/****************************************************************
 * An operand within a literal of a CNF. An operand is either a *
 * string or a variable with an attribute. May also be invalid  *
 * to return an error. In this case, isInvalid() returns true.  *
 ****************************************************************/


CNFOperand::CNFOperand() : m_type(OPERAND_TYPE_INVALID), m_variable(""),
        m_attribute(""), m_string("")
{
}


CNFOperand::~CNFOperand()
{
}


CNFOperand CNFOperand::fromInvalid()
{
    return CNFOperand();
}


CNFOperand CNFOperand::fromString(const QString& str)
{
    CNFOperand op;
    op.m_type = OPERAND_TYPE_STRING;
    op.m_string = str;
    return op;
}


CNFOperand CNFOperand::fromVariable(const QString& var,
        const QString& attr)
{
    CNFOperand op;
    op.m_type = OPERAND_TYPE_VAR;
    op.m_variable = var;
    op.m_attribute = attr;
    return op;
}


bool CNFOperand::isString() const
{
    return m_type == OPERAND_TYPE_STRING;
}


bool CNFOperand::isVariable() const
{
    return m_type == OPERAND_TYPE_VAR;
}


bool CNFOperand::isInvalid() const
{
    return m_type == OPERAND_TYPE_INVALID;
}


QString CNFOperand::string() const
{
    return m_string;
}


QString CNFOperand::variableName() const
{
    return m_variable;
}


QString CNFOperand::variableAttribute() const
{
    return m_attribute;
}


/***********************************************************************
 * A literal within a CNF disjunction. Either a variable and attribute *
 * returning a boolean value or a predicate and to operators.          *
 ***********************************************************************/

CNFLiteral::CNFLiteral(bool negated, const CNFOperand& left, const QString& op,
        const CNFOperand& right) : m_type(LITERAL_TYPE_TWO_PARAMTER_PREDICATE),
        m_negated(negated), m_operator(op), m_left(left), m_right(right)
{
}


CNFLiteral::CNFLiteral(bool negated, const CNFOperand& boolVariable) :
        m_type(LITERAL_TYPE_BOOL_VARIABLE), m_negated(negated),
        m_operator(""), m_left(boolVariable), m_right(CNFOperand::fromInvalid())
{
}


CNFLiteral CNFLiteral::fromInvalid()
{
    return CNFLiteral(false, CNFOperand::fromInvalid(), "invalid",
            CNFOperand::fromInvalid());
}


CNFLiteral::~CNFLiteral()
{
}


bool CNFLiteral::isNegated() const
{
    return m_negated;
}


bool CNFLiteral::isPredicate() const
{
    return m_type == LITERAL_TYPE_TWO_PARAMTER_PREDICATE;
}


bool CNFLiteral::isBoolVariable() const
{
    return m_type == LITERAL_TYPE_BOOL_VARIABLE;
}


CNFOperand CNFLiteral::left() const
{
    return m_left;
}


CNFOperand CNFLiteral::right() const
{
    return m_right;
}


QString CNFLiteral::op() const
{
    return m_operator;
}


CNFOperand CNFLiteral::boolVariable() const
{
    return m_left;
}


QSet<QString> CNFLiteral::variableNames() const
{
    QSet<QString> vars;
    if (m_left.isVariable())
        vars.insert(m_left.variableName());
    if (m_right.isVariable())
        vars.insert(m_right.variableName());
    return vars;
}


void CNFLiteral::setNegated(bool negated)
{
    m_negated = negated;
}



/**************************************************************************
 * A disjunction within a CNF. The disjunction stores a list of literals. *
 * In any disjunction of a CNF, at least one literal must be true if the  *
 * CNF should evaluate to true.                                           *
 **************************************************************************/

void CNFDisjunction::addLiteral(const CNFLiteral& literal)
{
    m_literals.append(literal);
}


QMultiHash<QString, int> CNFDisjunction::variableLocations() const
{
    QMultiHash<QString, int> hash;

    for (int i = 0; i < m_literals.length(); ++i)
    {
        int index = i;
        QSet<QString> varsInLiteral = m_literals.at(i).variableNames();
        QList<QString> varList = varsInLiteral.toList();
        for (int j = 0; j < varList.length(); ++j)
            hash.insert(varList.at(j), index);
    }
    return hash;
}


int CNFDisjunction::numLiterals() const
{
    return m_literals.length();
}


const CNFLiteral* CNFDisjunction::literalAt(int i) const
{
    if (i >= 0 && i < m_literals.length())
        return &m_literals.at(i);
    else
        return NULL;
}


/**********************************************************************
 * A conjunctive normal form. A CNF is a conjunction of disjunctions. *
 **********************************************************************/

void ConjunctiveNormalForm::clear()
{
    m_disjunctions.clear();
}


void ConjunctiveNormalForm::addDisjunction(const CNFDisjunction&
        disjunction)
{
    m_disjunctions.append(disjunction);
}


QMultiHash<QString, CNFLiteralIndex> ConjunctiveNormalForm::variableLocations() const
{
    QMultiHash<QString, CNFLiteralIndex> hash;

    for (int i = 0; i < m_disjunctions.length(); ++i)
    {
        int disjunctionIndex = i;
        // keep track which variable is used where
        QMultiHash<QString, int> varLocation = m_disjunctions.at(i).variableLocations();
        QList<QString> keys = varLocation.uniqueKeys();

        for (int j = 0; j < keys.length(); ++j)
        {
            QList<int> indices = varLocation.values(keys.at(j));
            for (int k = 0; k < indices.length(); ++k)
            {
                CNFLiteralIndex index(disjunctionIndex, indices.at(k));
                hash.insert(keys.at(j), index);
            }
        }
    }

    return hash;
}


CNFDisjunction* ConjunctiveNormalForm::lastDisjunction()
{
    if (!m_disjunctions.empty())
        return &m_disjunctions.last();
    else
        return NULL;
}


int ConjunctiveNormalForm::numDisjunctions() const
{
    return m_disjunctions.length();
}


const CNFDisjunction* ConjunctiveNormalForm::disjunctionAt(int i) const
{
    if (i >= 0 && i < m_disjunctions.length())
        return &m_disjunctions[i];
    else
        return NULL;
}


const CNFLiteral* ConjunctiveNormalForm::literalAt(CNFLiteralIndex index) const
{
    const CNFDisjunction* disj = disjunctionAt(index.disjunctionIndex());
    if (disj != NULL)
        return disj->literalAt(index.literalIndex());
    else
        return NULL;
}
