/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: constructcnf.cpp 2458 2011-08-03 09:21:52Z ahuesgen $
 */

#include "../../include/vhistfindlib/constructcnf.h"

#include <QtDebug>

using namespace vhist;

/********************************************************
 * Transform tree into a tree of conjuctive normal form *
 ********************************************************/


enum CNFLevel
{
    CNF_LEVEL_BELOW_AND,
    CNF_LEVEL_BELOW_OR,
    CNF_LEVEL_BELOW_LITERAL
};

SPExpressionTreeNode handleDeMorgan(SPExpressionTreeNode current,
        const QString& opAfter)
{
    SPExpressionTreeNode opNode(new ExpressionTreeNode());
    opNode->m_value = ExpressionValue::fromOperator(opAfter);

    SPExpressionTreeNode notNode1(new ExpressionTreeNode());
    notNode1->m_value = ExpressionValue::fromOperator("!");

    SPExpressionTreeNode notNode2(new ExpressionTreeNode());
    notNode2->m_value = ExpressionValue::fromOperator("!");

    notNode1->m_children.append(current->m_children[0]);
    notNode2->m_children.append(current->m_children[1]);
    opNode->m_children.append(notNode1);
    opNode->m_children.append(notNode2);

    return opNode;
}


SPExpressionTreeNode handleDistributive(SPExpressionTreeNode a,
        SPExpressionTreeNode b, SPExpressionTreeNode c)
/**
 * transforms tree structure (a && b) || c into (a || c) && (b || c)
 */
{
    SPExpressionTreeNode andNode(new ExpressionTreeNode());
    andNode->m_value = ExpressionValue::fromOperator("&&");

    SPExpressionTreeNode orNode1(new ExpressionTreeNode());
    orNode1->m_value = ExpressionValue::fromOperator("||");

    SPExpressionTreeNode orNode2(new ExpressionTreeNode());
    orNode2->m_value = ExpressionValue::fromOperator("||");

    orNode1->m_children.append(a);
    orNode1->m_children.append(c);

    orNode2->m_children.append(b);
    orNode2->m_children.append(c->cloneTree());

    andNode->m_children.append(orNode1);
    andNode->m_children.append(orNode2);

    return andNode;
}


bool transformChildren(SPExpressionTreeNode parent,
        SPExpressionTreeNode current, int currentId);

bool transformTreeNode(SPExpressionTreeNode grandParent,
        SPExpressionTreeNode parent, SPExpressionTreeNode current,
        int parentId, int currentId)
{
    if (parent->m_value.isOperator() && parent->m_value.operatorName() == "!")
    {
        if (current->m_value.isOperator() && current->m_value.operatorName() == "!")
        {
            // transform not not x into x, throw away parent and current
            grandParent->m_children.replace(parentId, current->m_children[0]);
            return true;
        }
        else if (current->m_value.isOperator() && current->m_value.operatorName().contains("&&"))
        {
            // transform not (x and y) into (not x) or (not y)
            grandParent->m_children.replace(parentId, handleDeMorgan(current, "||"));
            return true;
        }
        else if (current->m_value.isOperator() && current->m_value.operatorName().contains("||"))
        {
            // transform not (x or y) into (not x) and (not y)
            grandParent->m_children.replace(parentId, handleDeMorgan(current, "&&"));
            return true;
        }
    }
    else if (parent->m_value.isOperator() && parent->m_value.operatorName().contains("||"))
    {
        if (current->m_value.isOperator() && current->m_value.operatorName().contains("&&"))
        {
            // structure is either (a && b) || c or c || (a && b)
            SPExpressionTreeNode a = current->m_children[0];
            SPExpressionTreeNode b = current->m_children[1];
            SPExpressionTreeNode c;
            if (currentId == 0)
                c = parent->m_children[1];
            else
                c = parent->m_children[0];
            grandParent->m_children.replace(parentId, handleDistributive(a, b, c));
            return true;
        }
    }

    return transformChildren(parent, current, currentId);
}


bool transformChildren(SPExpressionTreeNode parent,
        SPExpressionTreeNode current, int currentId)
{
    bool modified = false;
    for (int i = 0; i < current->m_children.length(); ++i)
    {
        SPExpressionTreeNode child = current->m_children.at(i);
        bool subModified = transformTreeNode(parent, current, child, currentId, i);
        modified = modified || subModified;
    }
    return modified;
}


/*****************************************************
 * Transform a tree in CNF form into a CNF structure *
 *****************************************************/

CNFOperand makeOperand(const SPExpressionTreeNode current)
{
    ExpressionValue& v = current->m_value;

    if (v.isVariable())
        return CNFOperand::fromVariable(v.variableName(), v.variableAttribute());
    else if (v.isString())
        return CNFOperand::fromString(v.string());
    else
        return CNFOperand::fromInvalid();
}


CNFLiteral makeLiteral(const SPExpressionTreeNode current)
{
    if (current->m_value.isOperator() && current->m_value.operatorName() == "!")
    {
        CNFLiteral lit = makeLiteral(current->m_children[0]);
        lit.setNegated(true);
        return lit;
    }
    else if (current->m_value.isOperator())
    {
        CNFOperand left = makeOperand(current->m_children[0]);
        CNFOperand right = makeOperand(current->m_children[1]);
        QString op = current->m_value.operatorName();
        return CNFLiteral(false, left, op, right);
    }
    else if (current->m_value.isVariable())
    {
        CNFOperand boolVariable = makeOperand(current);
        return CNFLiteral(false, boolVariable);
    }
    else
        return CNFLiteral::fromInvalid();
}


void createCNFStructure(const SPExpressionTreeNode node,
        CNFLevel level, ConjunctiveNormalForm& cnf)
{
    ExpressionValue v = node->m_value;
    if (level == CNF_LEVEL_BELOW_AND &&
            (!v.isOperator() || !v.operatorName().contains("&&")))
    {
        CNFDisjunction disjunction;
        cnf.addDisjunction(disjunction);
        level = CNF_LEVEL_BELOW_OR;
    }

    if (level == CNF_LEVEL_BELOW_OR &&
            (!v.isOperator() || !v.operatorName().contains("||")))
    {
        CNFDisjunction* disjunction = cnf.lastDisjunction();
        CNFLiteral literal = makeLiteral(node);
        disjunction->addLiteral(literal);
        return;
    }

    for (int i = 0; i < node->m_children.length(); ++i)
        createCNFStructure(node->m_children.at(i), level, cnf);
}


ConjunctiveNormalForm vhist::constructCNF(const SPExpressionTreeNode tree)
{
    if (!tree)
        return ConjunctiveNormalForm();
    SPExpressionTreeNode copy = tree->cloneTree();
    SPExpressionTreeNode fakeParent(new ExpressionTreeNode());
    fakeParent->m_value = ExpressionValue::fromOperator("fakeRoot");
    fakeParent->m_children.append(copy);

    bool treeChanged = true;
    while (treeChanged)
    {
        copy = fakeParent->m_children[0];
        //qDebug() << copy;
        treeChanged = transformChildren(fakeParent, copy, 0);
    }

    ConjunctiveNormalForm cnf;
    createCNFStructure(copy, CNF_LEVEL_BELOW_AND, cnf);
    return cnf;
}
