/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: evaluatecnf.cpp 2169 2011-05-30 09:11:44Z ahuesgen $
 */

#include "../../include/vhistfindlib/evaluatecnf.h"
#include "../../include/unvhistlib/summary.h"
#include "../../include/vhistfindlib/expressiontreenode.h"
#include "../../include/vhistfindlib/returnvalue.h"
#include "../../include/vhistfindlib/evaloperators.h"

#include <QString>
#include <QList>
#include <QHash>

using namespace vhist;

enum EvaluationResult
{
    EVALUATION_RESULT_UNTESTED = 0,
    EVALUATION_RESULT_TRUE = 1,
    EVALUATION_RESULT_FALSE = 2
};


struct StaticParams
{
    QList<QString> variables;
    QMultiHash<QString, CNFLiteralIndex> varLocations;
    QList<QHash<QString, ReturnValue> > replacements;
    ConjunctiveNormalForm cnf;
};


struct DynamicParams
{
    QList<bool> replacementInUse;
    QList<int> variableSubstitutions;
    QList<QList<EvaluationResult> > cnfResult;
};

typedef ReturnValue(OperatorHandler2)(const ReturnValue&, const ReturnValue&);

QHash<QString, ReturnValue> convertStepToHash(const vhist::Workflowstep& wfs)
{
    QHash<QString, ReturnValue> wfsHash;
    wfsHash.insert("hash", ReturnValue::fromString(QString("wfs:%1").arg(wfs.tagID)));
    wfsHash.insert("tagid", ReturnValue::fromNumber(wfs.tagID));
    wfsHash.insert("vhistrootfile", ReturnValue::fromString(wfs.vhistRootFile));
    wfsHash.insert("timestamp", ReturnValue::fromString(wfs.timestamp));
    wfsHash.insert("title", ReturnValue::fromString(wfs.title));
    wfsHash.insert("description", ReturnValue::fromString(wfs.description));
    wfsHash.insert("comment", ReturnValue::fromString(wfs.comment));
    wfsHash.insert("command", ReturnValue::fromString(wfs.command));
    wfsHash.insert("tool", ReturnValue::fromString(wfs.tool));
    wfsHash.insert("toolpath", ReturnValue::fromString(wfs.toolPath));
    wfsHash.insert("host", ReturnValue::fromString(wfs.host));
    wfsHash.insert("user", ReturnValue::fromString(wfs.user));
    wfsHash.insert("creator", ReturnValue::fromString(wfs.creator));
    wfsHash.insert("iswfs", ReturnValue::fromBool(true));
    wfsHash.insert("isfile", ReturnValue::fromBool(false));
    QString attributes = wfs.userAttrOrder.join("\n").toLower();
    wfsHash.insert("userattrs", ReturnValue::fromString(attributes));

    for (int i = 0; i < wfs.userAttrOrder.length(); ++i)
    {
        QString key = wfs.userAttrOrder[i];
        QString value = wfs.userAttrs[key];
        wfsHash.insert("user:" + key.toLower(), ReturnValue::fromString(value));
    }

    QStringList keys = wfsHash.uniqueKeys();
    QStringList allValues;
    for (int i = 0; i < keys.length(); ++i)
    {
        ReturnValue v = wfsHash.value(keys.at(i), ReturnValue(RETURN_VALUE_INVALID_TYPE, ""));
        if (v.isString())
            allValues.append(v.toString());
    }
    wfsHash.insert("any", ReturnValue::fromString(allValues.join("\n")));

    return wfsHash;
}


QHash<QString, ReturnValue> convertFileToHash(const vhist::FileSummary& file,
        int parentId)
{
    QHash<QString, ReturnValue> fileHash;
    fileHash.insert("hash", ReturnValue::fromString(
            QString("file:%1:%2").arg(parentId).arg(file.tagID)));
    fileHash.insert("tagid", ReturnValue::fromNumber(file.tagID));
    fileHash.insert("filename", ReturnValue::fromString(file.filename));
    fileHash.insert("filepath", ReturnValue::fromString(file.filepath));
    fileHash.insert("filetype", ReturnValue::fromString(file.filetype));
    fileHash.insert("description", ReturnValue::fromString(file.description));
    fileHash.insert("comment", ReturnValue::fromString(file.comment));
    fileHash.insert("lastmodified", ReturnValue::fromString(file.lastModified));
    fileHash.insert("embedded", ReturnValue::fromBool(file.embedded));
    fileHash.insert("compressed", ReturnValue::fromBool(file.compressedMethod == "flate"));
    fileHash.insert("filepurpose", ReturnValue::fromString(file.filePurpose));
    if (file.filePurpose == "InFile")
    {
        fileHash.insert("isinfile", ReturnValue::fromBool(true));
        fileHash.insert("isoutfile", ReturnValue::fromBool(false));
    }
    else
    {
        fileHash.insert("isinfile", ReturnValue::fromBool(false));
        fileHash.insert("isoutfile", ReturnValue::fromBool(true));
    }
    fileHash.insert("filesize", ReturnValue::fromNumber(file.uncompressedSize));
    fileHash.insert("cfilesize", ReturnValue::fromNumber(file.compressedSize));
    fileHash.insert("md5", ReturnValue::fromString(file.uncompressedMD5));
    fileHash.insert("cmd5", ReturnValue::fromString(file.compressedMD5));
    fileHash.insert("iswfs", ReturnValue::fromBool(false));
    fileHash.insert("isfile", ReturnValue::fromBool(true));
    fileHash.insert("parentid", ReturnValue::fromNumber(parentId));
    QString attributes = file.userAttrOrder.join(", ").toLower();
    fileHash.insert("userattrs", ReturnValue::fromString(attributes));

    for (int i = 0; i < file.userAttrOrder.length(); ++i)
    {
        QString key = file.userAttrOrder[i];
        QString value = file.userAttrs[key];
        fileHash.insert("user:" + key, ReturnValue::fromString(value));
    }

    QStringList keys = fileHash.uniqueKeys();
    QStringList allValues;
    for (int i = 0; i < keys.length(); ++i)
    {
        ReturnValue v = fileHash.value(keys.at(i), ReturnValue(RETURN_VALUE_INVALID_TYPE, ""));
        if (v.isString())
            allValues.append(v.toString());
    }
    fileHash.insert("any", ReturnValue::fromString(allValues.join("\n")));

    return fileHash;
}


int findNextVariableIndex(const DynamicParams& dynamics)
{
    for (int i = 0; i < dynamics.variableSubstitutions.length(); ++i)
    {
        if (dynamics.variableSubstitutions.at(i) < 0)
            return i;
    }

    return -1;
}


ReturnValue operandToReturnValue(const StaticParams& statics,
        const DynamicParams& dynamics, const CNFOperand& operand)
{
    if (operand.isString())
        return ReturnValue::fromString(operand.string());
    else if (operand.isVariable())
    {
        int varIndex = statics.variables.indexOf(operand.variableName());
        int replacementIndex = dynamics.variableSubstitutions.at(varIndex);
        if (replacementIndex >= 0)
        {
            const QHash<QString, ReturnValue>& hash = statics.replacements.at(replacementIndex);
            QString attribute = operand.variableAttribute().toLower();
            ReturnValue value = hash.value(attribute,
                    ReturnValue(RETURN_VALUE_NOT_MATCHED, ""));
            return value;
        }
        else
            return ReturnValue(RETURN_VALUE_NOT_ASSIGNED_YET, "");
    }
    else
        return ReturnValue(RETURN_VALUE_UNKNOWN_TOKEN_TYPE, "");
}


EvaluationResult evaluatePredicate(const StaticParams& statics,
        const DynamicParams& dynamics, const CNFLiteral& literal)
{
    ReturnValue left = operandToReturnValue(statics, dynamics, literal.left());
    ReturnValue right = operandToReturnValue(statics, dynamics, literal.right());
    QString op = literal.op();

    if (left.isUnassignedVariable() || right.isUnassignedVariable())
        return EVALUATION_RESULT_UNTESTED;

    QHash<QString, OperatorHandler2*> operators;
    operators.insert("==", &evaluateEq);
    operators.insert("!=", &evaluateNEq);
    operators.insert("<", &evaluateLess);
    operators.insert("<=", &evaluateLessEq);
    operators.insert(">", &evaluateGreater);
    operators.insert(">=", &evaluateGreaterEq);
    operators.insert("%contains", &evaluateContains);
    operators.insert("%matches", &evaluateMatches);
    operators.insert("%isapprox", &evaluateIsApprox);

    OperatorHandler2* opHandler = operators.value(op, NULL);

    ReturnValue result = opHandler(left, right);
    if (result.isBool() && result.toBool() == true)
        return EVALUATION_RESULT_TRUE;
    else
        return EVALUATION_RESULT_FALSE;
}


EvaluationResult evaluateBoolVariable(const StaticParams& statics,
        const DynamicParams& dynamics, const CNFLiteral& literal)
{
    ReturnValue var = operandToReturnValue(statics, dynamics,
            literal.boolVariable());

    if (var.isUnassignedVariable())
        return EVALUATION_RESULT_UNTESTED;
    if (var.isBool() && var.toBool() == true)
        return EVALUATION_RESULT_TRUE;
    else
        return EVALUATION_RESULT_FALSE;
}


EvaluationResult evaluateLiteral(const StaticParams& statics,
        const DynamicParams& dynamics, const CNFLiteral& literal)
{
    EvaluationResult result;
    if (literal.isPredicate())
        result = evaluatePredicate(statics, dynamics, literal);
    else if (literal.isBoolVariable())
        result = evaluateBoolVariable(statics, dynamics, literal);
    else
        return EVALUATION_RESULT_FALSE;

    if (literal.isNegated() && result == EVALUATION_RESULT_FALSE)
        return EVALUATION_RESULT_TRUE;
    else if (literal.isNegated() && result == EVALUATION_RESULT_TRUE)
        return EVALUATION_RESULT_FALSE;
    else
        return result;
}


bool disjunctionIsSatisfiable(QList<EvaluationResult> disjunctionResults)
{
    for (int i = 0; i < disjunctionResults.length(); ++i)
        if (disjunctionResults.at(i) != EVALUATION_RESULT_FALSE)
            return true;
    return false;
}


QString getVariableValue(const QHash<QString, ReturnValue>& sectionOrFile)
{
    if (sectionOrFile.value("iswfs", ReturnValue::fromBool(false)).toBool())
    {
        QString title = sectionOrFile.value("title", ReturnValue::fromString("")).toString();
        return QString("workflowstep: %1").arg(title);
    }
    else if (sectionOrFile.value("isfile", ReturnValue::fromBool(false)).toBool())
    {
        QString title = sectionOrFile.value("filename", ReturnValue::fromString("")).toString();
        return QString("file: %1").arg(title);
    }
    else
        return QString("not a file of workflowstep ...");
}



QHash<QString, QString> getVariableAssociations(const StaticParams& statics,
        const DynamicParams& dynamics)
{
    QHash<QString, QString> result;
    for (int i = 0; i < statics.variables.length(); ++i)
    {
        QString variableName = statics.variables.at(i);
        if (!variableName.startsWith("@"))
        {
            int substitutionIndex = dynamics.variableSubstitutions.at(i);
            const QHash<QString, ReturnValue>& sub = statics.replacements.at(substitutionIndex);
            result.insert(variableName, getVariableValue(sub));
        }
    }
    return result;
}


bool davisPutnam(const StaticParams& statics, DynamicParams dynamics,
        QHash<QString, QString>& variableAssociations);

bool evaluateVariables(const StaticParams& statics, DynamicParams dynamics,
        int variableIndex, QHash<QString, QString>& variableAssociations)
{
    QString varName = statics.variables.at(variableIndex);
    QList<CNFLiteralIndex> locations = statics.varLocations.values(varName);

    for (int i = 0; i < locations.length(); ++i)
    {
        const CNFLiteralIndex& index = locations.at(i);
        const CNFLiteral* literal = statics.cnf.literalAt(index);
        EvaluationResult result = evaluateLiteral(statics, dynamics, *literal);
        QList<EvaluationResult>& disjunctionResult = dynamics.cnfResult[index.disjunctionIndex()];
        disjunctionResult[index.literalIndex()] = result;
        if (!disjunctionIsSatisfiable(disjunctionResult))
            return false;
    }
    return davisPutnam(statics, dynamics, variableAssociations);
}


bool davisPutnam(const StaticParams& statics, DynamicParams dynamics,
        QHash<QString, QString>& variableAssociations)
{
    int varIndex = findNextVariableIndex(dynamics);
    if (varIndex == -1) // no more variables
    {
        variableAssociations = getVariableAssociations(statics, dynamics);
        return true; // no more variables
    }

    QString varname = statics.variables.at(varIndex);
    for (int i = 0; i < statics.replacements.length(); ++i)
    {
        if (varname.startsWith("@"))
        {
            dynamics.variableSubstitutions[varIndex] = i;
            bool success = evaluateVariables(statics, dynamics, varIndex,
                    variableAssociations);
            if (success)
                return true;
        }
        else if (!dynamics.replacementInUse.at(i))
        {
            dynamics.variableSubstitutions[varIndex] = i;
            dynamics.replacementInUse[i] = true;
            bool success = evaluateVariables(statics, dynamics, varIndex,
                    variableAssociations);
            dynamics.replacementInUse[i] = false;
            if (success)
                return true;
        }
    }
    // no working substitution found
    return false;
}


bool vhist::evaluateCNF(const Summaries& summaries, const ConjunctiveNormalForm& cnf,
        QHash<QString, QString>& variableAssociations)
{
    // initialize static parameters
    StaticParams statics;

    // create hashtables for files and workflowsteps
    for (int i = 0; i < summaries.length(); ++i)
    {
        const vhist::Workflowstep& step = summaries.at(i).step;
        statics.replacements.append(convertStepToHash(step));

        for (int j = 0; j < step.files.length(); ++j)
            statics.replacements.append(convertFileToHash(step.files.at(j),
                    step.tagID));
    }

    statics.cnf = cnf;
    statics.varLocations = cnf.variableLocations();
    statics.variables = statics.varLocations.uniqueKeys();

    // initialize dynamic parameters
    int numVariable = statics.variables.length();
    int numReplacements = statics.replacements.length();

    DynamicParams dynamics;
    for (int i = 0; i < numVariable; ++i)
        dynamics.variableSubstitutions.append(-1);

    for (int i = 0; i < numReplacements; ++i)
        dynamics.replacementInUse.append(false);

    for (int i = 0; i < cnf.numDisjunctions(); ++i)
    {
        dynamics.cnfResult.append(QList<EvaluationResult>());
        const CNFDisjunction* disj = cnf.disjunctionAt(i);
        for (int j = 0; j < disj->numLiterals(); ++j)
            dynamics.cnfResult[i].append(EVALUATION_RESULT_UNTESTED);
    }

    return davisPutnam(statics, dynamics, variableAssociations);
}
