/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: parsequery.cpp 2458 2011-08-03 09:21:52Z ahuesgen $
 */

#include "../../include/vhistfindlib/parsequery.h"
#include "../../include/vhistfindlib/checkquerytypes.h"
#include "../../include/vhistfindlib/operatorinfo.h"

#include <QString>
#include <QStringList>
#include <QList>
#include <QStack>
#include <QRegExp>

#include <QtDebug>

using namespace vhist;


/**********************************************************************
 * Regular expressions for different token types (used by the lexer). *
 **********************************************************************/

static const QString regexWhitespaceStr("[ \t\n\r]+");

static const QString regexVariable1Str(
        "[a-zA-Z]*\\:user\\{([^}\\\\]|\\\\\\\\|\\\\\\})+\\}");
static const QString regexVariable2Str("[a-zA-Z]*\\:[a-zA-Z0-9]+(\\:[a-zA-Z0-9]+)?");

// unescaped string1: '((?:[^'\\]|\\\\|\\')*)'
static const QString regexString1Str("'((?:[^'\\\\]|\\\\\\\\|\\\\')*)'");
// unescaped string2: "((?:[^"\\]|\\\\|\\")*)"
static const QString regexString2Str("\"((?:[^\"\\\\]|\\\\\\\\|\\\\\")*)\"");
static const QString regexString3Str("[^:&|!%=><\\s\'\")(]+");

static const QString regexOperatorBinaryStr("&&|\\|\\||==|!=|<|<=|>|>=|%contains\\s|%contains$|%matches\\s|%matches$|%isapprox\\s|%isapprox$");
static const QString regexOperatorUnaryStr("\\!");
static const QString regexBracketsStr("\\(|\\)");


/***************************
 * Miscellaneous typedefs. *
 ***************************/

typedef QList<ExpressionValue> TokenStream;
typedef QList<TokenStream> TokenStreams;



/**************
 * Lexer code *
 **************/

ExpressionValue unescapeString1(const QString& string, int pos, const QString&
        originalText)
{
    QString s = string;
    s.replace("\\\\", "\\").replace("\\'", "'");
    return ExpressionValue::fromString(s, pos, originalText);
}


ExpressionValue unescapeString2(const QString& string, int pos, const QString&
        originalText)
{
    QString s = string;
    s.replace("\\\\", "\\").replace("\\\"", "\"");
    return ExpressionValue::fromString(s, pos, originalText);
}


ExpressionValue unescapeString3(const QString& string, int pos, const QString&
        originalText)
{
    return ExpressionValue::fromString(string, pos, originalText);
}


TokenStream lexQuery(const QString& text, QueryErrors& errors)
/**
 * Takes a \a text string and splits it up into individual tokens representing
 * strings, variables, operators and brackets. If a parsing error occures,
 * it is returned in the \a errors list.
 **/
{
    TokenStream tokens;

    // init regular expressions
    QRegExp regexWhitespace(regexWhitespaceStr);
    QRegExp regexVariable1(regexVariable1Str);
    QRegExp regexVariable2(regexVariable2Str);
    QRegExp regexString1(regexString1Str);
    QRegExp regexString2(regexString2Str);
    QRegExp regexString3(regexString3Str);
    QRegExp regexOperatorBinary(regexOperatorBinaryStr);
    QRegExp regexOperatorUnary(regexOperatorUnaryStr);
    QRegExp regexBrackets(regexBracketsStr);

    int offset = 0;
    while (offset < text.length())
    {
        // filter out whitespace
        if (regexWhitespace.indexIn(text, offset) == offset)
            offset += regexWhitespace.cap(0).length();
        // handle variables
        else if (regexVariable1.indexIn(text, offset) == offset)
        {
            QString captured = regexVariable1.cap(0);
            tokens.append(ExpressionValue::fromVariable(captured,
                    offset, captured));
            offset += captured.length();
        }
        else if (regexVariable2.indexIn(text, offset) == offset)
        {
            QString captured = regexVariable2.cap(0);
            tokens.append(ExpressionValue::fromVariable(captured,
                    offset, captured));
            offset += captured.length();
        }
        // handle strings
        else if (regexString1.indexIn(text, offset) == offset)
        {
            QString captured = regexString1.cap(0);
            tokens.append(unescapeString1(regexString1.cap(1), offset,
                    captured));
            offset += captured.length();
        }
        else if (regexString2.indexIn(text, offset) == offset)
        {
            QString captured = regexString2.cap(0);
            tokens.append(unescapeString2(regexString2.cap(1), offset,
                    captured));
            offset += captured.length();
        }
        else if (regexString3.indexIn(text, offset) == offset)
        {
            QString captured = regexString3.cap(0);
            tokens.append(unescapeString3(captured, offset, captured));
            offset += captured.length();
        }
        // handle operators
        else if (regexOperatorBinary.indexIn(text, offset) == offset)
        {
            QString captured = regexOperatorBinary.cap(0);
            QString op = captured.toLower().trimmed();
            tokens.append(ExpressionValue::fromOperator(op, offset, captured));
            offset += op.length();
        }
        else if (regexOperatorUnary.indexIn(text, offset) == offset)
        {
            QString captured = regexOperatorUnary.cap(0);
            QString op = captured.toLower().trimmed();
            tokens.append(ExpressionValue::fromOperator(op, offset, captured));
            offset += op.length();
        }
        // handle brackets
        else if (regexBrackets.indexIn(text, offset) == offset)
        {
            QString bracket = regexBrackets.cap(0);
            tokens.append(ExpressionValue::fromOperator(bracket, offset, bracket));
            offset += bracket.length();
        }
        else // nothing matched, must be an error ...
        {
            // generate error message
            QString errorMessage("Error while reading text at position %1 (%2):\n"
                    " input seems to be invalid");

            QString problematicText = text.mid(offset, 20);
            if (text.length() > offset + 20)
                problematicText += "...";

            QString e = errorMessage.arg(offset).arg(problematicText);
            errors.append(QueryError(e, QueryError::QUERY_ERROR_TYPE_ERROR,
                    offset, text.mid(offset)));
            return QList<ExpressionValue>();
        }
    }
    
    /*for (int i = 0; i < tokens.length(); ++i)
    {
        const ExpressionValue& token = tokens.at(i);
        qDebug() << token;
    }*/
    
    return tokens;
}


/*******************************************************
 * Syntax validity check and splitting of expressions. *
 *******************************************************/

ExpressionValue createImplizitAnd(int position)
{
    return ExpressionValue::fromOperator("implizit &&", position, "");
}

QPair<TokenStream, int> checkSubExpressionCorrectness(const TokenStream& tokens,
        int nextPos, bool endsWithBracket, QueryErrors& errors)
{
    TokenStream result;
    int i = nextPos;
    bool nextCanBeBinaryOperator = false;

    while (i < tokens.length())
    {
        // get next token
        const ExpressionValue& curToken = tokens.at(i);
        if (nextCanBeBinaryOperator)
        {
            // looking for a op b, current token is op => OK
            if (curToken.isOperator() && curToken.numArgs() == 2)
            {
                nextCanBeBinaryOperator = false;
                result.append(curToken);
            }
            // found a closing bracket
            else if (curToken.isOperator() && curToken.operatorName() == ")")
            {
                // closing bracket was expected?
                if (endsWithBracket)
                {
                    result.append(curToken);
                    return QPair<TokenStream, int>(result, i);
                }
                // too many closing brackets
                else
                {
                    QString errorMsg("Error at position %1:\n found a closing "
                            " bracket ) but did not except one.");
                    QString e = errorMsg.arg(curToken.position());
                    errors.append(QueryError(e,
                            QueryError::QUERY_ERROR_TYPE_ERROR,
                            curToken.position(), curToken.originalText()));
                    return QPair<TokenStream, int>(TokenStream(), -(i+1));
                }
            }
            // ALL POSSIBILITIES BELOW REQUIRE AN IMPLIZIT &&
            // found an opening bracket
            else if (curToken.isOperator() && curToken.operatorName() == "(")
            {
                result.append(createImplizitAnd(curToken.position()));
                result.append(curToken);

                QPair<TokenStream, int> subResult;
                subResult = checkSubExpressionCorrectness(tokens, i+1, true, errors);

                if (subResult.second < 0)
                    return subResult;
                else
                {
                    result.append(subResult.first);
                    i = subResult.second;
                }
            }
            // found a "not", next must not be a binary operator or a closing
            // bracket
            else if (curToken.isOperator() && curToken.operatorName() == "!")
            {
                nextCanBeBinaryOperator = false;
                result.append(createImplizitAnd(curToken.position()));
                result.append(curToken);
            }
            else // must be a string of a variable
            {
                result.append(createImplizitAnd(curToken.position()));
                result.append(curToken);
            }
        }
        else // next must be a variable, a string, a unary operator or an
             // opening bracket
        {
            if (!curToken.isOperator())
            {
                nextCanBeBinaryOperator = true;
                result.append(curToken);
            }
            else // is an operator (hopefully "(" or "!")
            {
                if (curToken.operatorName() == "(")
                {
                    result.append(curToken);

                    QPair<TokenStream, int> subResult;
                    subResult = checkSubExpressionCorrectness(tokens, i+1, true, errors);

                    if (subResult.second < 0)
                        return subResult;
                    else
                    {
                        result.append(subResult.first);
                        i = subResult.second;
                    }
                    nextCanBeBinaryOperator = true;
                }
                else if (curToken.operatorName() == "!")
                {
                    result.append(curToken);
                }
                else
                {
                    QString errorMsg("Error at position %1:\n expected a string, "
                            "a variable, an open bracket '(' or an unary "
                            "operator (e.g. !), but found '%2'.");
                    QString e = errorMsg.arg(curToken.position())
                            .arg(curToken.originalText());
                    errors.append(QueryError(e,
                            QueryError::QUERY_ERROR_TYPE_ERROR,
                            curToken.position(), curToken.originalText()));
                    return QPair<TokenStream, int>(TokenStream(), -(i+1)); // i may be zero...
                }
            }
        }
        i++;
    }

    // more error checks
    if (tokens.length() == 0)
    {
        QString errorMsg("Error after position 0:\n query is empty.");
        errors.append(QueryError(errorMsg, QueryError::QUERY_ERROR_TYPE_ERROR,
                0, ""));
        return QPair<TokenStream, int>(TokenStream(), -1);
    }

    if (endsWithBracket || !nextCanBeBinaryOperator)
    {
        QString errorMsg("Error after position %1:\n premature end of query "
                " after %2.");
        const ExpressionValue& token = tokens.at(tokens.length()-1);
        int posAfterToken = token.position() + token.length();
        QString e = errorMsg.arg(posAfterToken).arg(token.originalText());
        errors.append(QueryError(e, QueryError::QUERY_ERROR_TYPE_ERROR,
                token.position(), token.originalText()));
        return QPair<TokenStream, int>(TokenStream(), -(tokens.length()+1));
    }

    return QPair<TokenStream, int>(result, tokens.length());
}


TokenStream checkSyntax(const TokenStream& tokens, QueryErrors& errors)
{
    QPair<TokenStream, int> result = checkSubExpressionCorrectness(tokens, 0,
            false, errors);
    return result.first;
}


ExpressionValue makeEndOfStreamOperator()
{
    return ExpressionValue::fromOperator("EOS");
}


SPExpressionTreeNode createTree(QList<ExpressionValue> tokens,
        const QString& queryString, QueryErrors&)
{
    if (tokens.isEmpty())
        return SPExpressionTreeNode(NULL);
    
    QStack<SPExpressionTreeNode> nodeStack;
    QStack<ExpressionValue> operatorStack;
    int bracketPrecedence = 0;
    tokens.append(makeEndOfStreamOperator());

    for (int i = 0; i < tokens.length(); ++i)
    {
        ExpressionValue token = tokens[i];
        if (token.isVariable() || token.isString())
        {
            SPExpressionTreeNode node(new ExpressionTreeNode);
            node->m_value = token;
            node->m_pos = token.position();
            node->m_originalText = token.originalText();

            if (token.isVariable())
                node->m_variables.insert(token.variableName());
            nodeStack.push(node);
        }
        else if (token.isOperator())
        {
            if (token.operatorName() == "(")
                bracketPrecedence += 10;
            else if (token.operatorName() == ")")
                bracketPrecedence -= 10;
            else
            {
                token.setPrecedence(token.precedence() + bracketPrecedence);
                while (!operatorStack.isEmpty() &&
                    (token.precedence() < operatorStack.top().precedence() ||
                     (token.precedence() == operatorStack.top().precedence() &&
                     token.association() == 'l')))
                {
                    ExpressionValue nextOp = operatorStack.pop();
                    SPExpressionTreeNode node(new ExpressionTreeNode);
                    node->m_value = nextOp;

                    // finding left and right margin of query part
                    int left  = queryString.length();
                    int right = -1;

                    for (int c = 0; c < nextOp.numArgs(); ++c)
                    {
                        SPExpressionTreeNode child = nodeStack.pop();
                        node->m_children.prepend(child);
                        node->m_variables.unite(child->m_variables);
                        left  = qMin(left, child->m_pos);
                        right = qMax(right, child->m_pos + child->m_originalText.length());
                    }
                    node->m_originalText = queryString.mid(left, right - left);
                    node->m_pos = left;
                    nodeStack.push(node);
                }
                operatorStack.push(token);
            }
        }
    }
    
    SPExpressionTreeNode root = nodeStack.pop();
    //qDebug() << root;
    return root;
}


SPExpressionTreeNode makeAnyContainsNode(SPExpressionTreeNode stringNode)
/**
 * Creates a subtree for the expression ":any %contains <string>". The tree node
 * \a stringNode is recycled in the tree.
 */
{
    SPExpressionTreeNode anyNode = SPExpressionTreeNode(new ExpressionTreeNode);
    anyNode->m_pos = stringNode->m_pos;
    anyNode->m_originalText = "";
    anyNode->m_value = ExpressionValue::fromVariable(":any", stringNode->m_pos,
            "");

    SPExpressionTreeNode opNode = SPExpressionTreeNode(new ExpressionTreeNode);
    opNode->m_pos = stringNode->m_pos;
    opNode->m_originalText = stringNode->m_originalText;
    opNode->m_value = ExpressionValue::fromOperator("%contains",
            stringNode->m_pos, "");

    opNode->m_children.append(anyNode);
    opNode->m_children.append(stringNode);

    return opNode;
}


SPExpressionTreeNode expandImplicitTreeNodes(SPExpressionTreeNode root, bool isRoot)
/**
 * Iterates over the tree and searches for all strings with a logical operator
 * as parent (or no parent at all). These nodes are expanded to
 * :any %contains <String>. Example: Foo || Bar is expanded to
 * :any %contains Foo || :any %contains Bar.
 */
{
    if (root == NULL)
        return root;

    // special case: tree with only one string a root
    if (isRoot && root->m_value.isString())
        return makeAnyContainsNode(root);

    // normal case: node is logical operator
    if (root->m_value.isOperator())
    {
        QString opname = root->m_value.operatorName();
        if (opname.contains("&&") || opname.contains("||") || opname == "!")
        {
            for (int i = 0; i < root->m_children.length(); ++i)
            {
                SPExpressionTreeNode child = root->m_children.at(i);
                if (child->m_value.isString())
                    root->m_children.replace(i, makeAnyContainsNode(child));
            }
        }
    }

    // recurse
    for (int i = 0; i < root->m_children.length(); ++i)
        expandImplicitTreeNodes(root->m_children.at(i), false);
    return root;
}

QSet<QString> expandUnnamedVariables(SPExpressionTreeNode root, int numUnnamedVariables = 0)
{
    QSet<QString> result;
    if (root->m_value.isVariable() && root->m_value.variableName() == "")
    {
        numUnnamedVariables += 1;
        QString varName = QString("@unnamed-%1").arg(numUnnamedVariables);
        root->m_value.setVariableName(varName);
        root->m_variables.insert(varName);
        result.insert(varName);
        //qDebug() << "added unnamed variable" << varName;
    }
    else
    {
        for (int i = 0; i < root->m_children.length(); ++i)
        {
            QSet<QString> newChildVars = expandUnnamedVariables(
                    root->m_children.at(i), numUnnamedVariables);
            numUnnamedVariables += newChildVars.size();
            result.unite(newChildVars);
        }
        root->m_variables.unite(result);
    }

    return result;
}

SPExpressionTreeNode vhist::parseQuery(const QString& query, QueryErrors& errors)
/**
 * Parses a query and returns a parse tree.
 *
 * Parses a \a query written in the vhistfind query language and generates a
 * tree from this \a query. This tree can then be used to check if a VHIST file
 * matches the \a query. If the \a query contains any \a errors, these \a
 * errors are returned inside the \a errors parameter and \c NULL is returned.
 *
 * @param[in] query The query, which should be parsed.
 * @param[out] errors A list of parser errors and warnings generated while
 * parsing.
 * @return A tree representing the query or \c NULL if an error occured
 * during parsing.
 */
{
    // clean up
    errors.clear();

    // lexical analysis of text stream
    QList<ExpressionValue> tokens = lexQuery(query, errors);
    if (containsErrors(errors))
        return SPExpressionTreeNode(NULL);

    // check if a valid tree can be constructed from the lexer tokens and
    // split subqueries
    //TokenStreams streams = splitStreams(tokens, errors);
    QList<ExpressionValue> stream = checkSyntax(tokens, errors);
    if (containsErrors(errors))
        return SPExpressionTreeNode(NULL);

    // build tree representing the query
    //SPExpressionTreeNode root = createCompleteTree(streams, query, errors);
    SPExpressionTreeNode root = createTree(stream, query, errors);
    if (containsErrors(errors))
        return SPExpressionTreeNode(NULL);

    if (!checkQueryTypes(root, errors))
        return SPExpressionTreeNode(NULL);

    // post processing
    root = expandImplicitTreeNodes(root, true);

    expandUnnamedVariables(root, 0);
    //qDebug() << root;

    return root;
}
